from .serializer import\
    Serializer,\
    RegisterSerializer,\
    CreateSerializer,\
    GetDefaultSerializer,\
    DesCmd,\
    SerCmd
from .jsonserializer import JsonSerializer
from .textserializer import TextSerializer
from .pyserializer import PySerializer
from .jsserializer import JsSerializer



'''
 script to create
 - conecepts.py
 - __init__.py
 
 Bjoern Annighoefer 2023
'''

import os

if __name__ == "__main__":
    os.system('python -m eoq3conceptsgen.generatefromconceptscli -i gen/concepts.py.mako -o concepts.py')
    os.system('python -m eoq3conceptsgen.generatefromconceptscli -i gen/__init__.py.mako -o __init__.py')
    
    


'''
2019 Bjoern Annighoefer
'''
import os,shutil
from datetime import datetime

class Backupper():
    def __init__(self,backupSourceDirs=[],backupDestinationDir='./backup',maxBackups=10):
        self.backupSourceDirs = backupSourceDirs
        self.backupDestinationDir = backupDestinationDir
        self.maxBackups = maxBackups
        
        
    def CreateBackup(self):
        #create a new directory within the backup path
        #create the backup dir if it does not exist
        if(not os.path.isdir(self.backupDestinationDir)):
            os.makedirs(self.backupDestinationDir)
        #build log folder name from current date
        now = datetime.now() # current date and time
        backupBasename = now.strftime("%Y-%m-%d_%H_%M_%S")
        n = 0
        #check if there is already a backup for that second
        while(os.path.isdir(os.path.join(self.backupDestinationDir,"%s_%05d"%(backupBasename,n)))):
            n+=1
        self.currentBackupDir = os.path.join(self.backupDestinationDir,"%s_%05d"%(backupBasename,n))
        
        #remove old backups, if limit is exceeded
        existingBackupDirs = []
        for name in os.listdir(self.backupDestinationDir): 
            existingBackupDir = os.path.join (self.backupDestinationDir, name)   
            if os.path.isdir(existingBackupDir):
                existingBackupDirs.append(existingBackupDir)
                
        nExistingBackupDirs = len(existingBackupDirs)
        nEntriesToDelete = max([nExistingBackupDirs-self.maxBackups+1,0]) #+1 for the newly created one
        if 0 < nEntriesToDelete:
            existingBackupDirs = sorted(existingBackupDirs)
            for i in range(nEntriesToDelete):
                shutil.rmtree(existingBackupDirs[i])
            nExistingBackupDirs -= nEntriesToDelete
        
        #on for the current log
        os.makedirs(self.currentBackupDir)
        nExistingBackupDirs += 1;
        
        for backupSourceDir in self.backupSourceDirs:
            if(os.path.isdir(backupSourceDir)):
                backupDestDir = os.path.basename(backupSourceDir)
                backupDestPath = os.path.join(self.currentBackupDir,backupDestDir)
                shutil.copytree(backupSourceDir,backupDestPath)
            else:
                print("warning: Should backup %s, but path was not found."%(backupSourceDir))
        
        print('Backup ready! (path: %s)'%(self.currentBackupDir))


'''

2022 Bjoern Annighoefer
'''

from ..command import Evt

import traceback
from threading import Lock
#type checking
from typing import Callable, Any, List

class ObserverInfo:
    def __init__(self, callback:Callable[[List[Evt],Any,object],None], context:Any, sessionId:str):
        self.callback = callback #the callback function of the event listener
        self.context = context #a unique identifier or None
        self.sessionId = sessionId #set if this observer is linked to a session
        self.queuedEvts = [] #events for this observer postponed for sending
        self.evts = [] #events for this observer marked for sending
        
'''
    EventProvider
'''
class Observable:
    def __init__(self):
        self.observers = {} # callback -> event
        self.observerQueueMutex = Lock()
        
    def Observe(self, callback:Callable[[List[Evt],Any,object],None], context:Any=None, sessionId:str=None)->None: #by default register for all events
        observerInfo = ObserverInfo(callback,context,sessionId)
        self.observerQueueMutex.acquire()
        try:
            self.observers[(callback,context,sessionId)] = observerInfo
        finally:
            self.observerQueueMutex.release()
        
    def Unobserve(self, callback:Callable[[List[Evt],Any,object],None],context:Any=None, sessionId:str=None)->None:
        self.observerQueueMutex.acquire()
        try:
            self.observers.pop((callback,context,sessionId))
        finally:
            self.observerQueueMutex.release()


    '''
        INTERNAL EVENT PROCESSING
    '''

    def IsEventDesired(self, evt:Evt, sourceSessionId:str, sessionId:str)->bool:
        return True
        
    def NotifyObservers(self, evts:List[Evt], sourceSessionId:str)->None: #sends multiple events   
        #print("%s notifying %d observers"%(self,len(self.observers)))
        newEvts = evts.copy() #make a copy to ensure that the list is not changed outside when the event notification loop runs. 
        self.observerQueueMutex.acquire()
        for o in self.observers.values():
            filteredEvts = [e for e in newEvts if self.IsEventDesired(e,sourceSessionId,o.sessionId)]
            try:
                if(0<len(filteredEvts)):
                    o.callback(filteredEvts,o.context,self)
                    #print("%s called"%(o.callback))
            except:
                print("EvtProvider: Warning observer callback failed:")
                traceback.print_exc()
        self.observerQueueMutex.release()
            
    def PostponeObserverNotification(self, evts:List[Evt], sourceSessionId:str)->None: 
        ''' Prepare events for being send to the observers but do not send them right away, 
        but wait for being released.
        
        '''
        newEvts = evts.copy() #make a copy to ensure that the list is not changed outside when the event notification loop runs. 
        self.observerQueueMutex.acquire()
        #print("%s: Postponing %d events: %s"%(self, len(newEvts),[e.a[0] for e in newEvts] ))
        for o in self.observers.values():
            filteredEvts = [e for e in newEvts if self.IsEventDesired(e,sourceSessionId,o.sessionId)]
            if(0<len(filteredEvts)):
                #print("%s: Queuing %d events for %s"%(self, len(filteredEvts),o.sessionId))
                o.queuedEvts.append(filteredEvts)
        self.observerQueueMutex.release()      
        
    def ClearPostponedNotifications(self):
        '''Clears the stored notifications without sending them
        '''
        self.observerQueueMutex.acquire()
        for o in self.observers.values():
            o.queuedEvts.clear()
        self.observerQueueMutex.release()
        
    def ReleasePostponedNotifications(self):
        '''Moves all queued events to the sending queues
        '''
        self.observerQueueMutex.acquire()
        for o in self.observers.values():
            o.evts += o.queuedEvts
            o.queuedEvts.clear()
        self.observerQueueMutex.release()
        
        
    def SendQueuedAndReleasedNotifications(self):
        ''' Sends out all events that have been queued
        '''
        self.observerQueueMutex.acquire()
        for o in self.observers.values():
            for filteredEvts in o.evts:
                try:
                    #print("%s gets %d events:"%(str(o.sessionId), len(filteredEvts) ))
                    o.callback(filteredEvts,o.context,self)
                    #print("%s sending %s"%(str(o.context), str(filteredEvts) ))
                except Exception as e:
                    print("EvtProvider: Warning observer callback failed: %s"%(str(e)))
                    traceback.print_exc()
            o.evts.clear()
        self.observerQueueMutex.release()
        


'''
2019 Bjoern Annighoefer
'''



from .logger import Logger,DEFAULT_LOGGER_LEVELS

class ConsoleLogger(Logger):
    '''A logger printing to the console
    '''
    def __init__(self,activeLevels=DEFAULT_LOGGER_LEVELS.L2_WARNING):
        super().__init__(activeLevels)

        
    #@Override         
    def _Log(self,level,msg):
        print("%s: %s"%(level,msg))


'''
 Bjoern Annighoefer 2022
 
'''

from .history import History

class Transaction:
    ''' The class to store the transaction information of one domain interaction.
    
    '''
    def __init__(self,tid:int,roc:bool,cid:int):
        self.tid = tid
        self.roc = roc #is read-only command?
        self.session = None
        self.changes = [] #dictionary which collects changes until the transaction is finished
        self.nRealChanges:int = 0 #collects the number of changes reported to the outside
        self.historyStack = [] #there is always one history on the stack
        self.nextHistory = History() #the next history that is put on the stack. Can be used to initialize values before it becomes active
        self.lastHistory = None #remember the last removed history 
        self.isMuted = False
        self.cid = cid #the change ID of the transaction
        
    def StartHistory(self)->History:
        newHistory = self.nextHistory
        self.historyStack.append(newHistory) #add a new history layer
        self.nextHistory = History() #create a new history for the next call
        return newHistory
        
    def StopHistory(self):
        self.lastHistory = self.historyStack.pop() #remove uppermost history
        
    def GetCurrentHistory(self)->History:
        if(0<len(self.historyStack)):
            return self.historyStack[-1]
        else:
            return None
    
    def GetLastHistory(self)->History:
        return self.lastHistory
    
    def GetNextHistory(self)->History:
        return self.nextHistory


from timeit import default_timer as timer #used to time the command's execution time.
import csv
from math import inf

class BenchmarkEntry:
    def __init__(self):
        self.n = 0
        #self.samples = []
        self.min = inf;
        self.max = -inf;
        self.avg = 0.0;
        
        self.innerMin = inf;
        self.innerMax = -inf;
        self.innerAvg = 0.0;
        
        self.outerMin = inf;
        self.outerMax = -inf;
        self.outerAvg = 0.0;
        
    def AddSample(self,dur,outerDur,innerDur):
        #self.samples.append(sample) #this might be to slow
        self.min = min(self.min,dur)
        self.max = max(self.max,dur)
        self.avg = (self.avg * self.n + dur) / (self.n+1)
        #outer values
        self.outerMin = min(self.outerMin,outerDur)
        self.outerMax = max(self.outerMax,outerDur)
        self.outerAvg = (self.outerAvg * self.n + outerDur) / (self.n+1)
        #inner values
        self.innerMin = min(self.innerMin,innerDur)
        self.innerMax = max(self.innerMax,innerDur)
        self.innerAvg = (self.innerAvg * self.n + innerDur) / (self.n+1)
        #increase samples
        self.n = self.n+1


class Benchmark:
    def __init__(self,maxNesting=1000):
        self.messures = {}
        self.maxNesting = maxNesting
        self.timestamps = [0.0 for i in range(self.maxNesting)] #init array
        self.durations = [0.0 for i in range(self.maxNesting)] #init array
        self.currentNesting = 0
        self.lastNesting = 0
        
    def InitMessure(self,key):
        self.messures[key] = BenchmarkEntry()
        
    def AddSample(self,key,dur,outerDur,innerDur):
        messure = None
        if(key in self.messures):
            messure = self.messures[key]
        else:
            messure = BenchmarkEntry()
            self.messures[key] = messure
        #add the new entry
        messure.AddSample(dur,outerDur,innerDur)
        
    def TimeIt(self):
        return timer()
    
    def Start(self):
        t = self.TimeIt()
        self.lastNesting = self.currentNesting # track the history of start calls
        self.currentNesting += 1
        self.timestamps[self.lastNesting] = t
        self.durations[self.currentNesting] = 0.0 #reset the inner
        return t
        
    def Stop(self,key):
        t = self.TimeIt()
        self.lastNesting = self.currentNesting # track the history of start calls
        self.currentNesting += -1
        b = self.timestamps[self.currentNesting]
        outerDur = t-b
        #store the total inner duration for the next measure at this level, but only, if this is not the lowest level.
        if(self.currentNesting>0):
            self.durations[self.currentNesting] += outerDur 
        innerDur = 0.0
        # consider the inner duration if there where nested benchmark calls.
        if(self.lastNesting>self.currentNesting):
            innerDur = self.durations[self.lastNesting]
        dur = outerDur - innerDur
        self.AddSample(key, dur, outerDur, innerDur)
        return dur
        
    def Validate(self):
        if(self.currentNesting > 0):
            print('BENCHMARK WARNING: expected benchmark nesting to be 0, but got %d. Did you have unequal number of Start() and Stop() calls?'%(self.currentNesting))
    
    def SaveToFile(self,filename):
        self.Validate()
        # open the file in the write mode
        with open(filename, 'w',newline='') as f:
            # create the csv writer
            writer = csv.writer(f,delimiter=';',quotechar ='"',quoting=csv.QUOTE_MINIMAL)
            # write header
            writer.writerow(['key','n','min','max','avg','outerMin','outerMax','outerAvg','innerMin','innerMax','innerAvg'])
            # write measures
            for k,m in self.messures.items():
                writer.writerow([k,m.n,m.min,m.max,m.avg,m.outerMin,m.outerMax,m.outerAvg,m.innerMin,m.innerMax,m.innerAvg])
        



'''
Dealing with the interpretation of constraints and caching the results

2024 Bjoern Annighoefer

'''

from ..value import BOL, STR, LST
from ..query import Obj,Qry
from ..concepts import *
from ..config import Config, EOQ_DEFAULT_CONFIG
from ..serializer import CreateSerializer
from ..error import *
from ..mdb import ReadOnlyMdb

from .qryrunner import QryRunner

from typing import Dict, Union, Tuple

class ConstraintInfo:
    '''An entry in the constraints cache
    '''
    def __init__(self, constraint:Obj):
        self.constraint:Obj = constraint
        self.owner:Obj = None #the parent element of the constraint
        self.query:Qry = None
        self.targets:Dict[Obj,bool] = {} #the list of elements this constraint was evaluated for
        
class ConstraintEvaluationInfo:
    def __init__(self, constraint:Obj, target:Obj):
        self.constraint:Obj = constraint
        self.target:Obj = target #the element the constraint was evaluated for
        self.exception:str = None
        self.isValid = False
        self.evalCounter = 0 #how often was the constraint evaluated
        
class TargetInfo:
    def __init__(self, target:Obj):
        self.target:Obj = target
        self.evaluations:Dict[Obj,ConstraintEvaluationInfo] = {}#relates constraint and result

from ..verifier import Verifier

class ConstraintVerifier(Verifier):
    def __init__(self, config:Config = EOQ_DEFAULT_CONFIG):
        super().__init__()
        self.config = config
        #internals
        self.expressionSerializer = CreateSerializer(self.config.constraintSerializer)
        self.qryrunner = None #need a private qryrunner with full access is set later
        self.constraintsCache:Dict[Obj,ConstraintInfo] = {} #each constraint definition is stored here
        self.targetCache:Dict[Obj,TargetInfo] = {} #a combination of constraint and target forms an evaluation
        
    ### VERIFIER OVERRIDES ###
    #@override
    def Init(self, mdb:ReadOnlyMdb)->None:
        '''Initialize the Access controller and tell him the domain he is responsible for
        '''
        super().Init(mdb)
        self.qryrunner = QryRunner(self.mdb,self.config)
    
    #@override
    def CreatePreVerify(self, classId:Union[STR,Obj], createArgs:LST, target:Obj, recoveryArgs:LST, user:str=None)->None:
        '''Validate if a constraint contains a valid expression before it enters the mdb
        '''
        if(CONCEPTS.MXCONSTRAINT==classId and 2==len(createArgs)):
            expression = createArgs[1]
            if(STR == type(expression)): #if not STR, this create will fail anyhow
                self.expressionSerializer.DesQry(expression.GetVal()) #raises exception if it fails
    
    #@override
    def CreateNotify(self, classId:Union[STR,Obj], createArgs:LST, target:Obj, recoveryArgs:LST, user:str=None)->None:
        '''If a constraint has entered the mdb, add it to the cache
        '''
        if(CONCEPTS.MXCONSTRAINT==classId):
            # no further check necessary, since everything was validated beforehand
            constraint = target
            owner = createArgs[0]
            expression = createArgs[1]
            # create cache entry
            constraintInfo = ConstraintInfo(constraint)
            constraintInfo.owner = owner
            constraintInfo.query = self.expressionSerializer.DesQry(expression.GetVal())
            self.constraintsCache[constraint] = constraintInfo
                
    def DeleteNotify(self, target:Obj, classId:STR, createArgs:LST, recoveryArgs:LST, user:str=None)->None:
        '''Remove constraints from the cache if deleted
        '''
        # any deletion might cause a target being removed from the cache
        if(target in self.targetCache):
            targetInfo = self.targetCache[target]
            constraintInfo = self.constraintsCache[targetInfo.constraint]
            if(target in constraintInfo.targets):
                del constraintInfo.targets[target]
            del self.targetCache[target]
        # if a constraint itself is removed, remove its info and results
        if(CONCEPTS.MXCONSTRAINT==classId):
            c = target
            if(c in self.constraintsCache):
                constraintsInfo = self.constraintsCache[c]
                for t in constraintsInfo.targets.values():
                    if(t in self.targetCache):
                        targetInfo = self.targetCache[t]
                        if(c in targetInfo.evaluations):
                            del targetInfo.evaluations[c]
            del self.constraintsCache[c]
            
    ### CONSTRAINT METHODS ###
    def EvalConstraint(self, constraint:Obj, target:Obj)->Tuple[bool,str,bool]:
        '''evaluates a constraint for a given target and returns the result
        or raise an exception
        '''
        #get or create the evaluation info 
        constraintInfo = self.constraintsCache[constraint]
        targetInfo = None
        if(target in self.targetCache):
            targetInfo = self.targetCache[target]
        else: #create new target record
            targetInfo = TargetInfo(target)
            self.targetCache[target] = targetInfo
        evaluationInfo = None
        if(constraint in targetInfo.evaluations):
            evaluationInfo = targetInfo.evaluations[constraint]
        else: #create new evaluation record
            evaluationInfo = ConstraintEvaluationInfo(constraint,target)
            targetInfo.evaluations[constraint] = evaluationInfo #1. add to the results per target
            constraintInfo.targets[target] = True #2. add this as a target for the given constraint
        #store old values to obtain the changed text
        oldValid = evaluationInfo.isValid
        oldException = evaluationInfo.exception
        # run the evaluation
        try:
            result = self.qryrunner.EvalOnContextAndScope(target, constraintInfo.query, target, None)
            if(BOL==type(result)):
                evaluationInfo.exception = None
                evaluationInfo.isValid = result.GetVal()
            else:
                evaluationInfo.exception = EoqErrorAsString(EOQ_ERROR_INVALID_VALUE("No boolean result."))
                evaluationInfo.isValid = False
        except Exception as e:
            evaluationInfo.exception = EoqErrorAsString(EOQ_ERROR_RUNTIME("Evaluation failed: %s"%(str(e))))
            evaluationInfo.isValid = False
        #check if validation resulted in an update
        changed = (1>evaluationInfo.evalCounter) or (oldValid != evaluationInfo.isValid) or (oldException != evaluationInfo.exception)
        #increase counter
        evaluationInfo.evalCounter += 1
        #return
        return evaluationInfo.isValid, evaluationInfo.exception, changed
            
            
    
    


'''
 An EOQ protocol communication frame
 Bjoern Annighoefer 2019
 '''
 
from .. import __version__

class FRM_TYPE:
    '''There are five options a frame is used in regular EOQ communication:
    command, a result (of a command), events, as well as obs and ubs calls
    '''
    CMD = 'CMD'
    RES = 'RES'
    EVT = 'EVT'
    OBS = 'OBS'
    UBS = 'UBS'
        
class Frm:
    '''Frame a container for commands
    '''
    def __init__(self, typ:str, uid:int, ser:str, data:str, sid:str, roc:bool=False, version:str=__version__):
        ''' A frame transports a eoq commands (res and evt are also commands)
        '''
        self.ver:str = version #the version of eoq that created the frame
        self.typ:str = typ     #[3 chars] the type of communication (see above)
        self.uid:int = uid     #a unique identifier used to match requests and results
        self.ser:str = ser     #[3chars]the serialization type
        self.dat:str = data    #already serialized command
        self.sid:str = sid     #the session id
        self.roc:bool = roc    #is read-only command?
    def __repr__(self):
        return "Frm(typ=%s, ver=%s, uid=%s, sid=%s, ser=%s, roc=%d, dat=%s"%(self.typ,self.ver,self.uid,self.sid,self.ser,self.roc,self.dat)
    def __eq__(self,other):
        if(isinstance(other,Frm)):
            return (self.ver == other.ver and 
                    self.typ == other.typ and 
                    self.uid == other.uid and 
                    self.ser == other.ser and 
                    self.sid == other.sid and 
                    self.roc == other.roc and
                    self.dat == other.dat 
                    )
        else:
            return False



'''
PyEOQ Example: Change.py
----------------------------------------------------------------------------------------
Definition of the record to store local changes in the command runner. 

A change includes precise information about the change in oder to repeat it.
A change includes all information to revert the change.

2022 Bjoern Annighoefer
'''

from ..query import Obj
from ..value import VAL, PRM, BOL, U32, I64, STR, LST

from typing import Tuple

class Change:
    def __init__(self, postModError:bool=False, timestamp = -1, cid:int=-1, tid:int=-1, user:str='', sessionNumber:int = -1):
        self.postModError = postModError
        self.timestamp = timestamp
        self.cid = cid
        self.tid = tid, 
        self.user = user
        self.sessionNumber = sessionNumber
       
        
    def __repr__(self):
        return "%d/%d/%d(%s): %s(%d)"%(self.cid, self.tid, self.postModError, self.timestamp, self.user, self.sessionNumber)
        
class CrtChange(Change):
    def __init__(self, target:Obj, classId:STR, createArgs:LST \
                 , postModError:bool=False, timestamp = -1, cid:int=-1, tid:int=-1, user:str='', sessionNumber:int = -1):
        super().__init__(postModError, timestamp, cid, tid, user, sessionNumber)
        self.target = target
        self.classId = classId
        self.createArgs = createArgs
        
    def __repr__(self):
        return super().__repr__()+" CREATE %s %s -> %s"%(self.classId, self.createArgs, self.target)
    
class UpdChange(Change):
    def __init__(self, target:Obj, feature:STR, value:PRM, position:I64, oldValue:PRM, oldOwner:Obj, oldFeature:STR, oldIndex:I64\
                 , postModError:bool=False, timestamp = -1, cid:int=-1, tid:int=-1, user:str='', sessionNumber:int = -1):
        super().__init__(postModError, timestamp, cid, tid, user, sessionNumber)
        self.target = target
        self.feature = feature
        self.value = value
        self.position = position
        self.oldValue = oldValue
        self.oldOwner = oldOwner
        self.oldFeature = oldFeature
        self.oldIndex = oldIndex
        
    def __repr__(self):
        return super().__repr__()+" UPDATE %s %s %s %s"%(self.target, self.feature, self.value, self.position)
        
class DelChange(Change):
    def __init__(self, target:Obj, classId:Tuple[STR,Obj], createArgs:LST, recoveryArgs:LST\
                 , postModError:bool=False, timestamp = -1, cid:int=-1, tid:int=-1, user:str='', sessionNumber:int = -1):
        super().__init__(postModError, timestamp, cid, tid, user, sessionNumber)
        self.target = target
        self.classId = classId
        self.createArgs = createArgs
        self.recoveryArgs = recoveryArgs
        
    def __repr__(self):
        return super().__repr__()+" DELETE %s"%(self.target)
    
    
class ObsChange(Change):
    def __init__(self, evtType:str, evtKey:VAL\
                 , postModError:bool=False, timestamp=-1, cid:int=-1, tid:int=-1, user:str='', sessionNumber:int = -1):
        super().__init__(postModError, timestamp, cid, tid, user, sessionNumber)
        self.evtType = evtType
        self.evtKey = evtKey
        
    def __repr__(self):
        return super().__repr__()+" OBSERVE %s::%s"%(self.evtType, self.evtKey)
    
class UbsChange(Change):
    def __init__(self, evtType:str, evtKey:VAL, affectedSessionPublicId:str\
                 , postModError:bool=False, timestamp=-1, cid:int=-1, tid:int=-1, user:str='', sessionNumber:int = -1):
        super().__init__(postModError, timestamp, cid, tid, user, sessionNumber)
        self.evtType = evtType
        self.evtKey = evtKey
        self.affectedSessionPublicId = affectedSessionPublicId #to be stored, because ubs can be caused by other sessions
        
    def __repr__(self):
        return super().__repr__()+" UNOBSERVE %s::%s"%(self.evtType, self.evtKey)
    
class MsgChange(Change):
    def __init__(self, msgKey:STR, msg:STR\
                 , postModError:bool=False, timestamp=-1, cid:int=-1, tid:int=-1, user:str='', sessionNumber:int = -1):
        super().__init__(postModError, timestamp, cid, tid, user, sessionNumber)
        self.msgKey = msgKey
        self.msg = msg
        
    def __repr__(self):
        return super().__repr__()+" MESSAGE(%s): %s"%(self.msgKey, self.msg)
    
    
class VecChange(Change):
    def __init__(self, target:Obj,  constraint:Obj, isValid:BOL,  error:STR\
                 , postModError:bool=False, timestamp=-1, cid:int=-1, tid:int=-1, user:str='', sessionNumber:int = -1):
        super().__init__(postModError, timestamp, cid, tid, user, sessionNumber)
        self.target = target
        self.constraint = constraint
        self.isValid = isValid
        self.error = error
        
    def __repr__(self):
        return super().__repr__()+" VERIFICATION CHANGE: %s on %s %s %s"%(self.target, self.constraint, self.isValid, self.error)
    
class WatChange(Change):
    def __init__(self, wid:U32,  result:VAL,  error:STR\
                 , postModError:bool=False, timestamp=-1, cid:int=-1, tid:int=-1, user:str='', sessionNumber:int = -1):
        super().__init__(postModError, timestamp, cid, tid, user, sessionNumber)
        self.wid = wid
        self.result = result
        self.error = error
        
    def __repr__(self):
        return super().__repr__()+" WATCH: %s on %s %s %s"%(self.target, self.constraint, self.isValid, self.error)



# Generated from ../eoq3.g4 by ANTLR 4.13.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,27,182,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,2,9,7,9,2,10,7,10,2,11,7,11,2,12,7,12,2,13,7,13,
        2,14,7,14,2,15,7,15,2,16,7,16,2,17,7,17,2,18,7,18,2,19,7,19,2,20,
        7,20,2,21,7,21,2,22,7,22,2,23,7,23,2,24,7,24,2,25,7,25,2,26,7,26,
        1,0,1,0,1,0,5,0,58,8,0,10,0,12,0,61,9,0,1,0,3,0,64,8,0,1,1,5,1,67,
        8,1,10,1,12,1,70,9,1,1,1,1,1,1,1,5,1,75,8,1,10,1,12,1,78,9,1,1,1,
        3,1,81,8,1,1,2,1,2,1,3,1,3,1,3,1,3,1,4,1,4,1,4,3,4,92,8,4,1,5,1,
        5,1,5,1,5,1,5,1,5,1,5,1,5,1,5,1,5,3,5,104,8,5,1,6,1,6,1,7,1,7,1,
        8,1,8,1,9,1,9,1,10,1,10,1,11,1,11,1,12,1,12,1,13,1,13,3,13,122,8,
        13,1,14,1,14,1,15,1,15,1,16,1,16,1,17,1,17,1,18,1,18,3,18,134,8,
        18,1,18,1,18,5,18,138,8,18,10,18,12,18,141,9,18,1,18,1,18,1,19,1,
        19,1,19,5,19,148,8,19,10,19,12,19,151,9,19,1,19,1,19,1,20,1,20,3,
        20,157,8,20,1,20,1,20,5,20,161,8,20,10,20,12,20,164,9,20,1,21,1,
        21,3,21,168,8,21,1,22,1,22,3,22,172,8,22,1,23,1,23,1,24,1,24,1,25,
        1,25,1,26,1,26,1,26,0,0,27,0,2,4,6,8,10,12,14,16,18,20,22,24,26,
        28,30,32,34,36,38,40,42,44,46,48,50,52,0,1,2,0,13,13,16,18,178,0,
        54,1,0,0,0,2,68,1,0,0,0,4,82,1,0,0,0,6,84,1,0,0,0,8,91,1,0,0,0,10,
        103,1,0,0,0,12,105,1,0,0,0,14,107,1,0,0,0,16,109,1,0,0,0,18,111,
        1,0,0,0,20,113,1,0,0,0,22,115,1,0,0,0,24,117,1,0,0,0,26,121,1,0,
        0,0,28,123,1,0,0,0,30,125,1,0,0,0,32,127,1,0,0,0,34,129,1,0,0,0,
        36,131,1,0,0,0,38,144,1,0,0,0,40,154,1,0,0,0,42,167,1,0,0,0,44,171,
        1,0,0,0,46,173,1,0,0,0,48,175,1,0,0,0,50,177,1,0,0,0,52,179,1,0,
        0,0,54,59,3,2,1,0,55,56,5,15,0,0,56,58,3,2,1,0,57,55,1,0,0,0,58,
        61,1,0,0,0,59,57,1,0,0,0,59,60,1,0,0,0,60,63,1,0,0,0,61,59,1,0,0,
        0,62,64,5,15,0,0,63,62,1,0,0,0,63,64,1,0,0,0,64,1,1,0,0,0,65,67,
        5,12,0,0,66,65,1,0,0,0,67,70,1,0,0,0,68,66,1,0,0,0,68,69,1,0,0,0,
        69,71,1,0,0,0,70,68,1,0,0,0,71,76,5,13,0,0,72,73,5,14,0,0,73,75,
        3,4,2,0,74,72,1,0,0,0,75,78,1,0,0,0,76,74,1,0,0,0,76,77,1,0,0,0,
        77,80,1,0,0,0,78,76,1,0,0,0,79,81,3,6,3,0,80,79,1,0,0,0,80,81,1,
        0,0,0,81,3,1,0,0,0,82,83,3,8,4,0,83,5,1,0,0,0,84,85,5,1,0,0,85,86,
        3,46,23,0,86,87,3,26,13,0,87,7,1,0,0,0,88,92,3,10,5,0,89,92,3,38,
        19,0,90,92,3,36,18,0,91,88,1,0,0,0,91,89,1,0,0,0,91,90,1,0,0,0,92,
        9,1,0,0,0,93,104,3,12,6,0,94,104,3,14,7,0,95,104,3,16,8,0,96,104,
        3,18,9,0,97,104,3,20,10,0,98,104,3,22,11,0,99,104,3,24,12,0,100,
        104,3,32,16,0,101,104,3,34,17,0,102,104,3,26,13,0,103,93,1,0,0,0,
        103,94,1,0,0,0,103,95,1,0,0,0,103,96,1,0,0,0,103,97,1,0,0,0,103,
        98,1,0,0,0,103,99,1,0,0,0,103,100,1,0,0,0,103,101,1,0,0,0,103,102,
        1,0,0,0,104,11,1,0,0,0,105,106,5,20,0,0,106,13,1,0,0,0,107,108,5,
        21,0,0,108,15,1,0,0,0,109,110,5,22,0,0,110,17,1,0,0,0,111,112,5,
        23,0,0,112,19,1,0,0,0,113,114,5,24,0,0,114,21,1,0,0,0,115,116,5,
        25,0,0,116,23,1,0,0,0,117,118,5,26,0,0,118,25,1,0,0,0,119,122,3,
        28,14,0,120,122,3,30,15,0,121,119,1,0,0,0,121,120,1,0,0,0,122,27,
        1,0,0,0,123,124,7,0,0,0,124,29,1,0,0,0,125,126,5,8,0,0,126,31,1,
        0,0,0,127,128,5,27,0,0,128,33,1,0,0,0,129,130,5,19,0,0,130,35,1,
        0,0,0,131,133,5,2,0,0,132,134,3,8,4,0,133,132,1,0,0,0,133,134,1,
        0,0,0,134,139,1,0,0,0,135,136,5,3,0,0,136,138,3,8,4,0,137,135,1,
        0,0,0,138,141,1,0,0,0,139,137,1,0,0,0,139,140,1,0,0,0,140,142,1,
        0,0,0,141,139,1,0,0,0,142,143,5,4,0,0,143,37,1,0,0,0,144,145,5,5,
        0,0,145,149,3,40,20,0,146,148,3,40,20,0,147,146,1,0,0,0,148,151,
        1,0,0,0,149,147,1,0,0,0,149,150,1,0,0,0,150,152,1,0,0,0,151,149,
        1,0,0,0,152,153,5,6,0,0,153,39,1,0,0,0,154,156,3,42,21,0,155,157,
        3,52,26,0,156,155,1,0,0,0,156,157,1,0,0,0,157,162,1,0,0,0,158,159,
        5,7,0,0,159,161,3,52,26,0,160,158,1,0,0,0,161,164,1,0,0,0,162,160,
        1,0,0,0,162,163,1,0,0,0,163,41,1,0,0,0,164,162,1,0,0,0,165,168,3,
        44,22,0,166,168,3,50,25,0,167,165,1,0,0,0,167,166,1,0,0,0,168,43,
        1,0,0,0,169,172,3,46,23,0,170,172,3,48,24,0,171,169,1,0,0,0,171,
        170,1,0,0,0,172,45,1,0,0,0,173,174,5,10,0,0,174,47,1,0,0,0,175,176,
        5,11,0,0,176,49,1,0,0,0,177,178,5,9,0,0,178,51,1,0,0,0,179,180,3,
        8,4,0,180,53,1,0,0,0,15,59,63,68,76,80,91,103,121,133,139,149,156,
        162,167,171
    ]

class eoq3Parser ( Parser ):

    grammarFileName = "eoq3.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "' -> '", "'['", "','", "']'", "'('", 
                     "')'", "';'", "<INVALID>", "<INVALID>", "'$'", "<INVALID>", 
                     "'    '", "<INVALID>", "' '", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "'n0'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "STR_QUOTE", "QRY_FULL", "QRY_SHORT_HIS", "QRY_SHORT_OTH", 
                      "CMD_INDENT", "CMD_ID", "CMD_ARG_SEP", "CMD_SEP", 
                      "STR_SIMPLE_1", "STR_SIMPLE_2", "STR_SIMPLE_3", "NONE", 
                      "BOOL", "UINT32", "UINT64", "INT32", "INT64", "FLOAT32", 
                      "FLOAT64", "DATETIME" ]

    RULE_cmds = 0
    RULE_cmd = 1
    RULE_cmd_arg = 2
    RULE_cmd_res_name = 3
    RULE_val = 4
    RULE_val_prim = 5
    RULE_val_bol = 6
    RULE_val_u32 = 7
    RULE_val_u64 = 8
    RULE_val_i32 = 9
    RULE_val_i64 = 10
    RULE_val_f32 = 11
    RULE_val_f64 = 12
    RULE_val_str = 13
    RULE_val_str_simple = 14
    RULE_val_str_quote = 15
    RULE_val_dat = 16
    RULE_val_non = 17
    RULE_val_lst = 18
    RULE_val_qry = 19
    RULE_qry_seg = 20
    RULE_qry_sym = 21
    RULE_qry_sym_short = 22
    RULE_qry_sym_his = 23
    RULE_qry_sym_other = 24
    RULE_qry_sym_full = 25
    RULE_qry_arg = 26

    ruleNames =  [ "cmds", "cmd", "cmd_arg", "cmd_res_name", "val", "val_prim", 
                   "val_bol", "val_u32", "val_u64", "val_i32", "val_i64", 
                   "val_f32", "val_f64", "val_str", "val_str_simple", "val_str_quote", 
                   "val_dat", "val_non", "val_lst", "val_qry", "qry_seg", 
                   "qry_sym", "qry_sym_short", "qry_sym_his", "qry_sym_other", 
                   "qry_sym_full", "qry_arg" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    STR_QUOTE=8
    QRY_FULL=9
    QRY_SHORT_HIS=10
    QRY_SHORT_OTH=11
    CMD_INDENT=12
    CMD_ID=13
    CMD_ARG_SEP=14
    CMD_SEP=15
    STR_SIMPLE_1=16
    STR_SIMPLE_2=17
    STR_SIMPLE_3=18
    NONE=19
    BOOL=20
    UINT32=21
    UINT64=22
    INT32=23
    INT64=24
    FLOAT32=25
    FLOAT64=26
    DATETIME=27

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class CmdsContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def cmd(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(eoq3Parser.CmdContext)
            else:
                return self.getTypedRuleContext(eoq3Parser.CmdContext,i)


        def CMD_SEP(self, i:int=None):
            if i is None:
                return self.getTokens(eoq3Parser.CMD_SEP)
            else:
                return self.getToken(eoq3Parser.CMD_SEP, i)

        def getRuleIndex(self):
            return eoq3Parser.RULE_cmds

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCmds" ):
                listener.enterCmds(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCmds" ):
                listener.exitCmds(self)




    def cmds(self):

        localctx = eoq3Parser.CmdsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_cmds)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 54
            self.cmd()
            self.state = 59
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,0,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 55
                    self.match(eoq3Parser.CMD_SEP)
                    self.state = 56
                    self.cmd() 
                self.state = 61
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,0,self._ctx)

            self.state = 63
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==15:
                self.state = 62
                self.match(eoq3Parser.CMD_SEP)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CmdContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def CMD_ID(self):
            return self.getToken(eoq3Parser.CMD_ID, 0)

        def CMD_INDENT(self, i:int=None):
            if i is None:
                return self.getTokens(eoq3Parser.CMD_INDENT)
            else:
                return self.getToken(eoq3Parser.CMD_INDENT, i)

        def CMD_ARG_SEP(self, i:int=None):
            if i is None:
                return self.getTokens(eoq3Parser.CMD_ARG_SEP)
            else:
                return self.getToken(eoq3Parser.CMD_ARG_SEP, i)

        def cmd_arg(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(eoq3Parser.Cmd_argContext)
            else:
                return self.getTypedRuleContext(eoq3Parser.Cmd_argContext,i)


        def cmd_res_name(self):
            return self.getTypedRuleContext(eoq3Parser.Cmd_res_nameContext,0)


        def getRuleIndex(self):
            return eoq3Parser.RULE_cmd

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCmd" ):
                listener.enterCmd(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCmd" ):
                listener.exitCmd(self)




    def cmd(self):

        localctx = eoq3Parser.CmdContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_cmd)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 68
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==12:
                self.state = 65
                self.match(eoq3Parser.CMD_INDENT)
                self.state = 70
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 71
            self.match(eoq3Parser.CMD_ID)
            self.state = 76
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==14:
                self.state = 72
                self.match(eoq3Parser.CMD_ARG_SEP)
                self.state = 73
                self.cmd_arg()
                self.state = 78
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 80
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==1:
                self.state = 79
                self.cmd_res_name()


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Cmd_argContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def val(self):
            return self.getTypedRuleContext(eoq3Parser.ValContext,0)


        def getRuleIndex(self):
            return eoq3Parser.RULE_cmd_arg

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCmd_arg" ):
                listener.enterCmd_arg(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCmd_arg" ):
                listener.exitCmd_arg(self)




    def cmd_arg(self):

        localctx = eoq3Parser.Cmd_argContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_cmd_arg)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 82
            self.val()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Cmd_res_nameContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def qry_sym_his(self):
            return self.getTypedRuleContext(eoq3Parser.Qry_sym_hisContext,0)


        def val_str(self):
            return self.getTypedRuleContext(eoq3Parser.Val_strContext,0)


        def getRuleIndex(self):
            return eoq3Parser.RULE_cmd_res_name

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCmd_res_name" ):
                listener.enterCmd_res_name(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCmd_res_name" ):
                listener.exitCmd_res_name(self)




    def cmd_res_name(self):

        localctx = eoq3Parser.Cmd_res_nameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_cmd_res_name)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 84
            self.match(eoq3Parser.T__0)
            self.state = 85
            self.qry_sym_his()
            self.state = 86
            self.val_str()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ValContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def val_prim(self):
            return self.getTypedRuleContext(eoq3Parser.Val_primContext,0)


        def val_qry(self):
            return self.getTypedRuleContext(eoq3Parser.Val_qryContext,0)


        def val_lst(self):
            return self.getTypedRuleContext(eoq3Parser.Val_lstContext,0)


        def getRuleIndex(self):
            return eoq3Parser.RULE_val

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal" ):
                listener.enterVal(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal" ):
                listener.exitVal(self)




    def val(self):

        localctx = eoq3Parser.ValContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_val)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 91
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [8, 13, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27]:
                self.state = 88
                self.val_prim()
                pass
            elif token in [5]:
                self.state = 89
                self.val_qry()
                pass
            elif token in [2]:
                self.state = 90
                self.val_lst()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Val_primContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def val_bol(self):
            return self.getTypedRuleContext(eoq3Parser.Val_bolContext,0)


        def val_u32(self):
            return self.getTypedRuleContext(eoq3Parser.Val_u32Context,0)


        def val_u64(self):
            return self.getTypedRuleContext(eoq3Parser.Val_u64Context,0)


        def val_i32(self):
            return self.getTypedRuleContext(eoq3Parser.Val_i32Context,0)


        def val_i64(self):
            return self.getTypedRuleContext(eoq3Parser.Val_i64Context,0)


        def val_f32(self):
            return self.getTypedRuleContext(eoq3Parser.Val_f32Context,0)


        def val_f64(self):
            return self.getTypedRuleContext(eoq3Parser.Val_f64Context,0)


        def val_dat(self):
            return self.getTypedRuleContext(eoq3Parser.Val_datContext,0)


        def val_non(self):
            return self.getTypedRuleContext(eoq3Parser.Val_nonContext,0)


        def val_str(self):
            return self.getTypedRuleContext(eoq3Parser.Val_strContext,0)


        def getRuleIndex(self):
            return eoq3Parser.RULE_val_prim

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal_prim" ):
                listener.enterVal_prim(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal_prim" ):
                listener.exitVal_prim(self)




    def val_prim(self):

        localctx = eoq3Parser.Val_primContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_val_prim)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 103
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [20]:
                self.state = 93
                self.val_bol()
                pass
            elif token in [21]:
                self.state = 94
                self.val_u32()
                pass
            elif token in [22]:
                self.state = 95
                self.val_u64()
                pass
            elif token in [23]:
                self.state = 96
                self.val_i32()
                pass
            elif token in [24]:
                self.state = 97
                self.val_i64()
                pass
            elif token in [25]:
                self.state = 98
                self.val_f32()
                pass
            elif token in [26]:
                self.state = 99
                self.val_f64()
                pass
            elif token in [27]:
                self.state = 100
                self.val_dat()
                pass
            elif token in [19]:
                self.state = 101
                self.val_non()
                pass
            elif token in [8, 13, 16, 17, 18]:
                self.state = 102
                self.val_str()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Val_bolContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def BOOL(self):
            return self.getToken(eoq3Parser.BOOL, 0)

        def getRuleIndex(self):
            return eoq3Parser.RULE_val_bol

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal_bol" ):
                listener.enterVal_bol(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal_bol" ):
                listener.exitVal_bol(self)




    def val_bol(self):

        localctx = eoq3Parser.Val_bolContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_val_bol)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 105
            self.match(eoq3Parser.BOOL)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Val_u32Context(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def UINT32(self):
            return self.getToken(eoq3Parser.UINT32, 0)

        def getRuleIndex(self):
            return eoq3Parser.RULE_val_u32

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal_u32" ):
                listener.enterVal_u32(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal_u32" ):
                listener.exitVal_u32(self)




    def val_u32(self):

        localctx = eoq3Parser.Val_u32Context(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_val_u32)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 107
            self.match(eoq3Parser.UINT32)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Val_u64Context(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def UINT64(self):
            return self.getToken(eoq3Parser.UINT64, 0)

        def getRuleIndex(self):
            return eoq3Parser.RULE_val_u64

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal_u64" ):
                listener.enterVal_u64(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal_u64" ):
                listener.exitVal_u64(self)




    def val_u64(self):

        localctx = eoq3Parser.Val_u64Context(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_val_u64)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 109
            self.match(eoq3Parser.UINT64)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Val_i32Context(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INT32(self):
            return self.getToken(eoq3Parser.INT32, 0)

        def getRuleIndex(self):
            return eoq3Parser.RULE_val_i32

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal_i32" ):
                listener.enterVal_i32(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal_i32" ):
                listener.exitVal_i32(self)




    def val_i32(self):

        localctx = eoq3Parser.Val_i32Context(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_val_i32)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 111
            self.match(eoq3Parser.INT32)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Val_i64Context(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INT64(self):
            return self.getToken(eoq3Parser.INT64, 0)

        def getRuleIndex(self):
            return eoq3Parser.RULE_val_i64

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal_i64" ):
                listener.enterVal_i64(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal_i64" ):
                listener.exitVal_i64(self)




    def val_i64(self):

        localctx = eoq3Parser.Val_i64Context(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_val_i64)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 113
            self.match(eoq3Parser.INT64)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Val_f32Context(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def FLOAT32(self):
            return self.getToken(eoq3Parser.FLOAT32, 0)

        def getRuleIndex(self):
            return eoq3Parser.RULE_val_f32

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal_f32" ):
                listener.enterVal_f32(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal_f32" ):
                listener.exitVal_f32(self)




    def val_f32(self):

        localctx = eoq3Parser.Val_f32Context(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_val_f32)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 115
            self.match(eoq3Parser.FLOAT32)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Val_f64Context(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def FLOAT64(self):
            return self.getToken(eoq3Parser.FLOAT64, 0)

        def getRuleIndex(self):
            return eoq3Parser.RULE_val_f64

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal_f64" ):
                listener.enterVal_f64(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal_f64" ):
                listener.exitVal_f64(self)




    def val_f64(self):

        localctx = eoq3Parser.Val_f64Context(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_val_f64)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 117
            self.match(eoq3Parser.FLOAT64)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Val_strContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def val_str_simple(self):
            return self.getTypedRuleContext(eoq3Parser.Val_str_simpleContext,0)


        def val_str_quote(self):
            return self.getTypedRuleContext(eoq3Parser.Val_str_quoteContext,0)


        def getRuleIndex(self):
            return eoq3Parser.RULE_val_str

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal_str" ):
                listener.enterVal_str(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal_str" ):
                listener.exitVal_str(self)




    def val_str(self):

        localctx = eoq3Parser.Val_strContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_val_str)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 121
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [13, 16, 17, 18]:
                self.state = 119
                self.val_str_simple()
                pass
            elif token in [8]:
                self.state = 120
                self.val_str_quote()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Val_str_simpleContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STR_SIMPLE_1(self):
            return self.getToken(eoq3Parser.STR_SIMPLE_1, 0)

        def STR_SIMPLE_2(self):
            return self.getToken(eoq3Parser.STR_SIMPLE_2, 0)

        def STR_SIMPLE_3(self):
            return self.getToken(eoq3Parser.STR_SIMPLE_3, 0)

        def CMD_ID(self):
            return self.getToken(eoq3Parser.CMD_ID, 0)

        def getRuleIndex(self):
            return eoq3Parser.RULE_val_str_simple

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal_str_simple" ):
                listener.enterVal_str_simple(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal_str_simple" ):
                listener.exitVal_str_simple(self)




    def val_str_simple(self):

        localctx = eoq3Parser.Val_str_simpleContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_val_str_simple)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 123
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 466944) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Val_str_quoteContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STR_QUOTE(self):
            return self.getToken(eoq3Parser.STR_QUOTE, 0)

        def getRuleIndex(self):
            return eoq3Parser.RULE_val_str_quote

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal_str_quote" ):
                listener.enterVal_str_quote(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal_str_quote" ):
                listener.exitVal_str_quote(self)




    def val_str_quote(self):

        localctx = eoq3Parser.Val_str_quoteContext(self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_val_str_quote)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 125
            self.match(eoq3Parser.STR_QUOTE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Val_datContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DATETIME(self):
            return self.getToken(eoq3Parser.DATETIME, 0)

        def getRuleIndex(self):
            return eoq3Parser.RULE_val_dat

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal_dat" ):
                listener.enterVal_dat(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal_dat" ):
                listener.exitVal_dat(self)




    def val_dat(self):

        localctx = eoq3Parser.Val_datContext(self, self._ctx, self.state)
        self.enterRule(localctx, 32, self.RULE_val_dat)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 127
            self.match(eoq3Parser.DATETIME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Val_nonContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NONE(self):
            return self.getToken(eoq3Parser.NONE, 0)

        def getRuleIndex(self):
            return eoq3Parser.RULE_val_non

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal_non" ):
                listener.enterVal_non(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal_non" ):
                listener.exitVal_non(self)




    def val_non(self):

        localctx = eoq3Parser.Val_nonContext(self, self._ctx, self.state)
        self.enterRule(localctx, 34, self.RULE_val_non)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 129
            self.match(eoq3Parser.NONE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Val_lstContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def val(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(eoq3Parser.ValContext)
            else:
                return self.getTypedRuleContext(eoq3Parser.ValContext,i)


        def getRuleIndex(self):
            return eoq3Parser.RULE_val_lst

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal_lst" ):
                listener.enterVal_lst(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal_lst" ):
                listener.exitVal_lst(self)




    def val_lst(self):

        localctx = eoq3Parser.Val_lstContext(self, self._ctx, self.state)
        self.enterRule(localctx, 36, self.RULE_val_lst)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 131
            self.match(eoq3Parser.T__1)
            self.state = 133
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 268378404) != 0):
                self.state = 132
                self.val()


            self.state = 139
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==3:
                self.state = 135
                self.match(eoq3Parser.T__2)
                self.state = 136
                self.val()
                self.state = 141
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 142
            self.match(eoq3Parser.T__3)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Val_qryContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def qry_seg(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(eoq3Parser.Qry_segContext)
            else:
                return self.getTypedRuleContext(eoq3Parser.Qry_segContext,i)


        def getRuleIndex(self):
            return eoq3Parser.RULE_val_qry

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVal_qry" ):
                listener.enterVal_qry(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVal_qry" ):
                listener.exitVal_qry(self)




    def val_qry(self):

        localctx = eoq3Parser.Val_qryContext(self, self._ctx, self.state)
        self.enterRule(localctx, 38, self.RULE_val_qry)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 144
            self.match(eoq3Parser.T__4)
            self.state = 145
            self.qry_seg()
            self.state = 149
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 3584) != 0):
                self.state = 146
                self.qry_seg()
                self.state = 151
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 152
            self.match(eoq3Parser.T__5)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Qry_segContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def qry_sym(self):
            return self.getTypedRuleContext(eoq3Parser.Qry_symContext,0)


        def qry_arg(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(eoq3Parser.Qry_argContext)
            else:
                return self.getTypedRuleContext(eoq3Parser.Qry_argContext,i)


        def getRuleIndex(self):
            return eoq3Parser.RULE_qry_seg

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterQry_seg" ):
                listener.enterQry_seg(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitQry_seg" ):
                listener.exitQry_seg(self)




    def qry_seg(self):

        localctx = eoq3Parser.Qry_segContext(self, self._ctx, self.state)
        self.enterRule(localctx, 40, self.RULE_qry_seg)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 154
            self.qry_sym()
            self.state = 156
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 268378404) != 0):
                self.state = 155
                self.qry_arg()


            self.state = 162
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==7:
                self.state = 158
                self.match(eoq3Parser.T__6)
                self.state = 159
                self.qry_arg()
                self.state = 164
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Qry_symContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def qry_sym_short(self):
            return self.getTypedRuleContext(eoq3Parser.Qry_sym_shortContext,0)


        def qry_sym_full(self):
            return self.getTypedRuleContext(eoq3Parser.Qry_sym_fullContext,0)


        def getRuleIndex(self):
            return eoq3Parser.RULE_qry_sym

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterQry_sym" ):
                listener.enterQry_sym(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitQry_sym" ):
                listener.exitQry_sym(self)




    def qry_sym(self):

        localctx = eoq3Parser.Qry_symContext(self, self._ctx, self.state)
        self.enterRule(localctx, 42, self.RULE_qry_sym)
        try:
            self.state = 167
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [10, 11]:
                self.enterOuterAlt(localctx, 1)
                self.state = 165
                self.qry_sym_short()
                pass
            elif token in [9]:
                self.enterOuterAlt(localctx, 2)
                self.state = 166
                self.qry_sym_full()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Qry_sym_shortContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def qry_sym_his(self):
            return self.getTypedRuleContext(eoq3Parser.Qry_sym_hisContext,0)


        def qry_sym_other(self):
            return self.getTypedRuleContext(eoq3Parser.Qry_sym_otherContext,0)


        def getRuleIndex(self):
            return eoq3Parser.RULE_qry_sym_short

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterQry_sym_short" ):
                listener.enterQry_sym_short(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitQry_sym_short" ):
                listener.exitQry_sym_short(self)




    def qry_sym_short(self):

        localctx = eoq3Parser.Qry_sym_shortContext(self, self._ctx, self.state)
        self.enterRule(localctx, 44, self.RULE_qry_sym_short)
        try:
            self.state = 171
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [10]:
                self.enterOuterAlt(localctx, 1)
                self.state = 169
                self.qry_sym_his()
                pass
            elif token in [11]:
                self.enterOuterAlt(localctx, 2)
                self.state = 170
                self.qry_sym_other()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Qry_sym_hisContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def QRY_SHORT_HIS(self):
            return self.getToken(eoq3Parser.QRY_SHORT_HIS, 0)

        def getRuleIndex(self):
            return eoq3Parser.RULE_qry_sym_his

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterQry_sym_his" ):
                listener.enterQry_sym_his(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitQry_sym_his" ):
                listener.exitQry_sym_his(self)




    def qry_sym_his(self):

        localctx = eoq3Parser.Qry_sym_hisContext(self, self._ctx, self.state)
        self.enterRule(localctx, 46, self.RULE_qry_sym_his)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 173
            self.match(eoq3Parser.QRY_SHORT_HIS)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Qry_sym_otherContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def QRY_SHORT_OTH(self):
            return self.getToken(eoq3Parser.QRY_SHORT_OTH, 0)

        def getRuleIndex(self):
            return eoq3Parser.RULE_qry_sym_other

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterQry_sym_other" ):
                listener.enterQry_sym_other(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitQry_sym_other" ):
                listener.exitQry_sym_other(self)




    def qry_sym_other(self):

        localctx = eoq3Parser.Qry_sym_otherContext(self, self._ctx, self.state)
        self.enterRule(localctx, 48, self.RULE_qry_sym_other)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 175
            self.match(eoq3Parser.QRY_SHORT_OTH)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Qry_sym_fullContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def QRY_FULL(self):
            return self.getToken(eoq3Parser.QRY_FULL, 0)

        def getRuleIndex(self):
            return eoq3Parser.RULE_qry_sym_full

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterQry_sym_full" ):
                listener.enterQry_sym_full(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitQry_sym_full" ):
                listener.exitQry_sym_full(self)




    def qry_sym_full(self):

        localctx = eoq3Parser.Qry_sym_fullContext(self, self._ctx, self.state)
        self.enterRule(localctx, 50, self.RULE_qry_sym_full)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 177
            self.match(eoq3Parser.QRY_FULL)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Qry_argContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def val(self):
            return self.getTypedRuleContext(eoq3Parser.ValContext,0)


        def getRuleIndex(self):
            return eoq3Parser.RULE_qry_arg

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterQry_arg" ):
                listener.enterQry_arg(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitQry_arg" ):
                listener.exitQry_arg(self)




    def qry_arg(self):

        localctx = eoq3Parser.Qry_argContext(self, self._ctx, self.state)
        self.enterRule(localctx, 52, self.RULE_qry_arg)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 179
            self.val()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








from .value import\
    VAL_TYPES,\
    VAL,\
    RegisterValType,\
    SIV,\
    PRM,\
    NUM,\
    INT,\
    FLO,\
    BOL,\
    U32,\
    U64,\
    I32,\
    I64,\
    F32,\
    F64,\
    STR,\
    DAT,\
    QRY,\
    NON,\
    LST,\
    EncVal,\
    InitValOrNon,\
    ValidateVal,\
    ValidateValAndCast,\
    ValCompare,\
    ValFactory



'''
 deleter.py recursive delete function for a model tree based on concepts. 
 
 This file is generated form: cloner.py.mako on 2024-07-10 09:58:41.752063 by calling E:/1_ILS/Projects/2018_EOQ/Groups/PYEOQ/pyeoq/eoq3conceptsgen/generatefromconceptscli.py -i cloner.py.mako -o ../cloner.py.
 
 Bjoern Annighoefer 2023
'''

from ..concepts import CONCEPTS, MXMDB ,MXELEMENT ,MXCONSTRAINT ,M2PRIMITIVES ,M2PACKAGE ,M2MODEL ,M2ENUM ,M2OPTIONOFENUM ,M2CLASS ,M2ATTRIBUTE ,M2ASSOCIATION ,M2COMPOSITION ,M2INHERITANCE ,M1MODEL ,M1OBJECT ,M1FEATURE ,M1ATTRIBUTE ,M1ASSOCIATION ,M1COMPOSITION
from ..value import VAL, STR, NON, LST, I64
from ..query import Obj
from ..command import CLO_MODES
from ..error import EOQ_ERROR_INVALID_VALUE, EOQ_ERROR_INVALID_TYPE, EOQ_ERROR_INVALID_OPERATION

from .conceptwalker import ConceptWalker

from typing import Tuple, Any, Dict, List

class CloneInfo:
    def __init__(self, original:Obj, concept:STR, createArgNames:LST, createArgs:LST):
        self.original:Obj = original
        self.concept:str = concept
        self.createArgNames:LST = createArgNames
        self.createArgs:LST = createArgs
        self.clone = None

class CloneFulRun:
    '''This class is used to maintain global information in one delete call
    '''
    def __init__(self):
        self.elementsToClone:Dict[int,Obj] = {} #a dict sorting elements by order
        self.cloneInfoLut:Dict[Obj,CloneInfo] = {} #a look-up table for clone information

class Cloner(ConceptWalker):
    def __init__(self, cmdrunner, logger):
        super().__init__(cmdrunner)
        self.logger = logger
        #createArgs get handlers
        self.handlers = {}
        self.handlers[CONCEPTS.MXCONSTRAINT] = (self.__GetCreateArgsMxConstraint,self.__CopyBasePropertiesMxConstraint)
        self.handlers[CONCEPTS.M2PACKAGE] = (self.__GetCreateArgsM2Package,self.__CopyBasePropertiesM2Package)
        self.handlers[CONCEPTS.M2MODEL] = (self.__GetCreateArgsM2Model,self.__CopyBasePropertiesM2Model)
        self.handlers[CONCEPTS.M2ENUM] = (self.__GetCreateArgsM2Enum,self.__CopyBasePropertiesM2Enum)
        self.handlers[CONCEPTS.M2OPTIONOFENUM] = (self.__GetCreateArgsM2OptionOfEnum,self.__CopyBasePropertiesM2OptionOfEnum)
        self.handlers[CONCEPTS.M2CLASS] = (self.__GetCreateArgsM2Class,self.__CopyBasePropertiesM2Class)
        self.handlers[CONCEPTS.M2ATTRIBUTE] = (self.__GetCreateArgsM2Attribute,self.__CopyBasePropertiesM2Attribute)
        self.handlers[CONCEPTS.M2ASSOCIATION] = (self.__GetCreateArgsM2Association,self.__CopyBasePropertiesM2Association)
        self.handlers[CONCEPTS.M2COMPOSITION] = (self.__GetCreateArgsM2Composition,self.__CopyBasePropertiesM2Composition)
        self.handlers[CONCEPTS.M2INHERITANCE] = (self.__GetCreateArgsM2Inheritance,self.__CopyBasePropertiesM2Inheritance)
        self.handlers[CONCEPTS.M1MODEL] = (self.__GetCreateArgsM1Model,self.__CopyBasePropertiesM1Model)
        self.handlers[CONCEPTS.M1OBJECT] = (self.__GetCreateArgsM1Object,self.__CopyBasePropertiesM1Object)
        self.handlers[CONCEPTS.M1ATTRIBUTE] = (self.__GetCreateArgsM1Attribute,self.__CopyBasePropertiesM1Attribute)
        self.handlers[CONCEPTS.M1ASSOCIATION] = (self.__GetCreateArgsM1Association,self.__CopyBasePropertiesM1Association)
        self.handlers[CONCEPTS.M1COMPOSITION] = (self.__GetCreateArgsM1Composition,self.__CopyBasePropertiesM1Composition)

    def Clone(self, original:Obj, mode:STR, createArgOverrides:LST, tid:int)->Obj:
        ''' Clones an element according to the given mode and create arg overrides
        This is the main entry point into the function
        '''
        concept = self.__CheckCanClone(tid, original)
        self.__ValidateCreateArgOverrides(createArgOverrides)
        clone = NON()
        if(CLO_MODES.MIN==mode):
            clone = self.__CloneMinimum(tid, original,concept,createArgOverrides)
        elif(CLO_MODES.BAS==mode):
            clone = self.__CloneBase(tid, original,concept,createArgOverrides)
        elif(CLO_MODES.FUL==mode):
            clone = self.__CloneFull(tid, original,concept,createArgOverrides)
        return clone
    
    # CLONE MODE IMPLEMENTASIONS
    
    def __CheckCanClone(self, tid:int, original:Obj)->str:
        concept = self.cmdrunner._MdbRead(tid,original,STR(MXELEMENT.CONCEPT))
        if( concept not in self.handlers):
            raise EOQ_ERROR_INVALID_OPERATION("Cannot clone %s"%(concept))
        return concept.GetVal()
        
    def __CloneMinimum(self, tid:int, original:Obj, concept:str, createArgOverrides:LST)->Obj:
        '''Implements clone mode MIN
        '''
        (createArgNames,createArgs) = self.__GetCreateInfo(tid,original,concept)
        cloneCreateArgs = self.__OverrideCreateArgs(concept,createArgNames,createArgs,createArgOverrides)
        clone = self.cmdrunner._MdbCreate(tid,STR(concept),cloneCreateArgs)
        return clone
    
    def __CloneBase(self, tid:int, original:Obj, concept:str, createArgOverrides:LST)->Obj:
        clone = self.__CloneMinimum(tid,original,concept,createArgOverrides)
        copyBasePropertiesHandler = self.handlers[concept][1] # No try necessary, because checked before 
        copyBasePropertiesHandler(tid,original,clone)
        return clone
    
    def __CloneFull(self, tid:int, original:Obj, concept:str, createArgOverrides:LST)->Obj:
        clone = NON()
        cloneRun = CloneFulRun()
        # 1 collect all elements to be cloned
        self.Walk(tid,original,None,-1,cloneRun)
        # 2 arrange elements such that they can be cloned in that order without 
        # violating dependencies and remove the root element, which was already cloned.
        conflictfreeCloneChain = [e for k,e in sorted(cloneRun.elementsToClone.items())]
        # 4 cloning elements from the elementsToClone is straight forward
        for o in conflictfreeCloneChain:
            originalInfo = cloneRun.cloneInfoLut[o]
            internalOverrides = self.__CreateArgOverrides(originalInfo,cloneRun.cloneInfoLut)
            if(o == original):
                self.__UpdateCreateArgOverwrites(internalOverrides,createArgOverrides)
                clone = self.__CloneBase(tid,original,concept,internalOverrides)
                cloneRun.cloneInfoLut[original].clone = clone
            else:
                originalInfo.clone = self.__CloneBase(tid,o,originalInfo.concept,internalOverrides)
        return clone
    
    # AUXILLIARY METHODS
    
    def __UpdateCreateArgOverwrites(self,updatedOverwrites:LST, newOverwrites:LST):
        nNewOverwrites = len(newOverwrites)
        nUpdatedOverwrites = len(updatedOverwrites)
        for i in range(0,nNewOverwrites,2):
            matchFound = False
            for j in range(0,nUpdatedOverwrites,2):
                if(updatedOverwrites[j] == newOverwrites[i]):
                    updatedOverwrites[j+1] = newOverwrites[i+1]
                    matchFound = True
                    break
            if(not matchFound): 
                #no match was found add the values
                updatedOverwrites.append(newOverwrites[i])
                updatedOverwrites.append(newOverwrites[i+1])
    
    def __MdbUpdate(self, tid:int, target:Obj, featureName:STR, value:VAL, pos:I64):
        self.cmdrunner._MdbUpdate(tid,target,featureName,value,pos)
            
    def __ValidateCreateArgOverrides(self, createArgOverrides:LST)->bool:
        nOverrides = len(createArgOverrides)
        if(0 != nOverrides % 2):
            raise EOQ_ERROR_INVALID_VALUE("Number of entries in createArgOverrides must be even.")
        for i in range(0,nOverrides,2):
            if(STR != type(createArgOverrides[i])):
                raise EOQ_ERROR_INVALID_TYPE("createArgOverrides[%d] must be string."%(i))
        return True
    
    def __GetCreateInfo(self, tid:int, original:Obj, concept:str)->Tuple[STR,List[STR],List[VAL]]:
        ''' Figures out the concept type and then calls the specific 
        function to return the create arg names and values for this concept type
        '''
        createArgsGetHandler = self.handlers[concept][0] #cannot fail, because concept was checked before
        (createArgNames,createArgs) = createArgsGetHandler(tid,original)
        return (createArgNames,createArgs)
    
    def __OverrideCreateArgs(self, concept:STR, createArgNames:List[STR], createArgs:List[VAL], createArgOverrides:LST)->LST:
        ''' This replaces values in an createArgs list 
        according to a given createArgOverrites list.
        It returns the new createArgs with the replaced values
        '''
        updatedCreateArgs = list(createArgs)
        for i in range(0,len(createArgOverrides),2):
            overrideName = createArgOverrides[i]
            try:
                j = createArgNames.index(overrideName)
                updatedCreateArgs[j] = createArgOverrides[i+1]
            except ValueError:
                raise EOQ_ERROR_INVALID_VALUE("%s is no create arg of %s."%(overrideName,concept))
        return LST(updatedCreateArgs)
    
    def __CreateArgOverrides(self, originalInfo:CloneInfo, cloneInfoLut:Dict[Obj,CloneInfo])->LST:
        '''Creates the createArgOverrides list from all elements cloned so far
        '''
        createArgOverrides = LST([])
        for i in range(len(originalInfo.createArgNames)):
            o = originalInfo.createArgs[i]
            if(isinstance(o,Obj)):
                try:
                    c = cloneInfoLut[o].clone
                    createArgOverrides.append(originalInfo.createArgNames[i])
                    createArgOverrides.append(c)
                except KeyError:
                    pass # no need to replace this argument, it has no clone
        return createArgOverrides;

    # ## CONCEPT WALKER SPECIALIZATION

    #@override
    def _OnEnterMxConstraint(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        '''Collections all concepts (not super concepts) to build a causal chain for cloning
        '''
        if(target != successor and target not in data.cloneInfoLut):
            #only return elements if this is not a super concept walk
            (createArgNames,createArgs) = self.__GetCreateArgsMxConstraint(tid,target)
            order = target.v[0] #TODO: implement official method for order
            data.cloneInfoLut[target] = CloneInfo(target,CONCEPTS.MXCONSTRAINT,createArgNames,createArgs)
            data.elementsToClone[order] = target
    #@override
    def _OnEnterM2Package(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        '''Collections all concepts (not super concepts) to build a causal chain for cloning
        '''
        if(target != successor and target not in data.cloneInfoLut):
            #only return elements if this is not a super concept walk
            (createArgNames,createArgs) = self.__GetCreateArgsM2Package(tid,target)
            order = target.v[0] #TODO: implement official method for order
            data.cloneInfoLut[target] = CloneInfo(target,CONCEPTS.M2PACKAGE,createArgNames,createArgs)
            data.elementsToClone[order] = target
    #@override
    def _OnEnterM2Model(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        '''Collections all concepts (not super concepts) to build a causal chain for cloning
        '''
        if(target != successor and target not in data.cloneInfoLut):
            #only return elements if this is not a super concept walk
            (createArgNames,createArgs) = self.__GetCreateArgsM2Model(tid,target)
            order = target.v[0] #TODO: implement official method for order
            data.cloneInfoLut[target] = CloneInfo(target,CONCEPTS.M2MODEL,createArgNames,createArgs)
            data.elementsToClone[order] = target
    #@override
    def _OnEnterM2Enum(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        '''Collections all concepts (not super concepts) to build a causal chain for cloning
        '''
        if(target != successor and target not in data.cloneInfoLut):
            #only return elements if this is not a super concept walk
            (createArgNames,createArgs) = self.__GetCreateArgsM2Enum(tid,target)
            order = target.v[0] #TODO: implement official method for order
            data.cloneInfoLut[target] = CloneInfo(target,CONCEPTS.M2ENUM,createArgNames,createArgs)
            data.elementsToClone[order] = target
    #@override
    def _OnEnterM2OptionOfEnum(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        '''Collections all concepts (not super concepts) to build a causal chain for cloning
        '''
        if(target != successor and target not in data.cloneInfoLut):
            #only return elements if this is not a super concept walk
            (createArgNames,createArgs) = self.__GetCreateArgsM2OptionOfEnum(tid,target)
            order = target.v[0] #TODO: implement official method for order
            data.cloneInfoLut[target] = CloneInfo(target,CONCEPTS.M2OPTIONOFENUM,createArgNames,createArgs)
            data.elementsToClone[order] = target
    #@override
    def _OnEnterM2Class(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        '''Collections all concepts (not super concepts) to build a causal chain for cloning
        '''
        if(target != successor and target not in data.cloneInfoLut):
            #only return elements if this is not a super concept walk
            (createArgNames,createArgs) = self.__GetCreateArgsM2Class(tid,target)
            order = target.v[0] #TODO: implement official method for order
            data.cloneInfoLut[target] = CloneInfo(target,CONCEPTS.M2CLASS,createArgNames,createArgs)
            data.elementsToClone[order] = target
    #@override
    def _OnEnterM2Attribute(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        '''Collections all concepts (not super concepts) to build a causal chain for cloning
        '''
        if(target != successor and target not in data.cloneInfoLut):
            #only return elements if this is not a super concept walk
            (createArgNames,createArgs) = self.__GetCreateArgsM2Attribute(tid,target)
            order = target.v[0] #TODO: implement official method for order
            data.cloneInfoLut[target] = CloneInfo(target,CONCEPTS.M2ATTRIBUTE,createArgNames,createArgs)
            data.elementsToClone[order] = target
    #@override
    def _OnEnterM2Association(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        '''Collections all concepts (not super concepts) to build a causal chain for cloning
        '''
        if(target != successor and target not in data.cloneInfoLut):
            #only return elements if this is not a super concept walk
            (createArgNames,createArgs) = self.__GetCreateArgsM2Association(tid,target)
            order = target.v[0] #TODO: implement official method for order
            data.cloneInfoLut[target] = CloneInfo(target,CONCEPTS.M2ASSOCIATION,createArgNames,createArgs)
            data.elementsToClone[order] = target
    #@override
    def _OnEnterM2Composition(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        '''Collections all concepts (not super concepts) to build a causal chain for cloning
        '''
        if(target != successor and target not in data.cloneInfoLut):
            #only return elements if this is not a super concept walk
            (createArgNames,createArgs) = self.__GetCreateArgsM2Composition(tid,target)
            order = target.v[0] #TODO: implement official method for order
            data.cloneInfoLut[target] = CloneInfo(target,CONCEPTS.M2COMPOSITION,createArgNames,createArgs)
            data.elementsToClone[order] = target
    #@override
    def _OnEnterM2Inheritance(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        '''Collections all concepts (not super concepts) to build a causal chain for cloning
        '''
        if(target != successor and target not in data.cloneInfoLut):
            #only return elements if this is not a super concept walk
            (createArgNames,createArgs) = self.__GetCreateArgsM2Inheritance(tid,target)
            order = target.v[0] #TODO: implement official method for order
            data.cloneInfoLut[target] = CloneInfo(target,CONCEPTS.M2INHERITANCE,createArgNames,createArgs)
            data.elementsToClone[order] = target
    #@override
    def _OnEnterM1Model(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        '''Collections all concepts (not super concepts) to build a causal chain for cloning
        '''
        if(target != successor and target not in data.cloneInfoLut):
            #only return elements if this is not a super concept walk
            (createArgNames,createArgs) = self.__GetCreateArgsM1Model(tid,target)
            order = target.v[0] #TODO: implement official method for order
            data.cloneInfoLut[target] = CloneInfo(target,CONCEPTS.M1MODEL,createArgNames,createArgs)
            data.elementsToClone[order] = target
    #@override
    def _OnEnterM1Object(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        '''Collections all concepts (not super concepts) to build a causal chain for cloning
        '''
        if(target != successor and target not in data.cloneInfoLut):
            #only return elements if this is not a super concept walk
            (createArgNames,createArgs) = self.__GetCreateArgsM1Object(tid,target)
            order = target.v[0] #TODO: implement official method for order
            data.cloneInfoLut[target] = CloneInfo(target,CONCEPTS.M1OBJECT,createArgNames,createArgs)
            data.elementsToClone[order] = target
    #@override
    def _OnEnterM1Attribute(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        '''Collections all concepts (not super concepts) to build a causal chain for cloning
        '''
        if(target != successor and target not in data.cloneInfoLut):
            #only return elements if this is not a super concept walk
            (createArgNames,createArgs) = self.__GetCreateArgsM1Attribute(tid,target)
            order = target.v[0] #TODO: implement official method for order
            data.cloneInfoLut[target] = CloneInfo(target,CONCEPTS.M1ATTRIBUTE,createArgNames,createArgs)
            data.elementsToClone[order] = target
    #@override
    def _OnEnterM1Association(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        '''Collections all concepts (not super concepts) to build a causal chain for cloning
        '''
        if(target != successor and target not in data.cloneInfoLut):
            #only return elements if this is not a super concept walk
            (createArgNames,createArgs) = self.__GetCreateArgsM1Association(tid,target)
            order = target.v[0] #TODO: implement official method for order
            data.cloneInfoLut[target] = CloneInfo(target,CONCEPTS.M1ASSOCIATION,createArgNames,createArgs)
            data.elementsToClone[order] = target
    #@override
    def _OnEnterM1Composition(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any):
        '''Collections all concepts (not super concepts) to build a causal chain for cloning
        '''
        if(target != successor and target not in data.cloneInfoLut):
            #only return elements if this is not a super concept walk
            (createArgNames,createArgs) = self.__GetCreateArgsM1Composition(tid,target)
            order = target.v[0] #TODO: implement official method for order
            data.cloneInfoLut[target] = CloneInfo(target,CONCEPTS.M1COMPOSITION,createArgNames,createArgs)
            data.elementsToClone[order] = target
    # ## GET CREATE ARG HANDLERS

    def __GetCreateArgsMxConstraint(self, tid:int, original:Obj)->Tuple[List[STR],List[VAL]]:
        pElement = self.cmdrunner._MdbRead(tid,original,STR(MXCONSTRAINT.ELEMENT))
        pExpression = self.cmdrunner._MdbRead(tid,original,STR(MXCONSTRAINT.EXPRESSION))
        createArgNames = [STR(MXCONSTRAINT.ELEMENT), STR(MXCONSTRAINT.EXPRESSION)]
        createArgs = [pElement, pExpression]
        return (createArgNames,createArgs)

    def __GetCreateArgsM2Package(self, tid:int, original:Obj)->Tuple[List[STR],List[VAL]]:
        pName = self.cmdrunner._MdbRead(tid,original,STR(M2PACKAGE.NAME))
        pSuperpackage = self.cmdrunner._MdbRead(tid,original,STR(M2PACKAGE.SUPERPACKAGE))
        createArgNames = [STR(M2PACKAGE.NAME), STR(M2PACKAGE.SUPERPACKAGE)]
        createArgs = [pName, pSuperpackage]
        return (createArgNames,createArgs)

    def __GetCreateArgsM2Model(self, tid:int, original:Obj)->Tuple[List[STR],List[VAL]]:
        pName = self.cmdrunner._MdbRead(tid,original,STR(M2PACKAGE.NAME))
        createArgNames = [STR(M2PACKAGE.NAME)]
        createArgs = [pName]
        return (createArgNames,createArgs)

    def __GetCreateArgsM2Enum(self, tid:int, original:Obj)->Tuple[List[STR],List[VAL]]:
        pName = self.cmdrunner._MdbRead(tid,original,STR(M2ENUM.NAME))
        pPackage = self.cmdrunner._MdbRead(tid,original,STR(M2ENUM.PACKAGE))
        createArgNames = [STR(M2ENUM.NAME), STR(M2ENUM.PACKAGE)]
        createArgs = [pName, pPackage]
        return (createArgNames,createArgs)

    def __GetCreateArgsM2OptionOfEnum(self, tid:int, original:Obj)->Tuple[List[STR],List[VAL]]:
        pName = self.cmdrunner._MdbRead(tid,original,STR(M2OPTIONOFENUM.NAME))
        pValue = self.cmdrunner._MdbRead(tid,original,STR(M2OPTIONOFENUM.VALUE))
        pEnum = self.cmdrunner._MdbRead(tid,original,STR(M2OPTIONOFENUM.ENUM))
        createArgNames = [STR(M2OPTIONOFENUM.NAME), STR(M2OPTIONOFENUM.VALUE), STR(M2OPTIONOFENUM.ENUM)]
        createArgs = [pName, pValue, pEnum]
        return (createArgNames,createArgs)

    def __GetCreateArgsM2Class(self, tid:int, original:Obj)->Tuple[List[STR],List[VAL]]:
        pName = self.cmdrunner._MdbRead(tid,original,STR(M2CLASS.NAME))
        pIsabstract = self.cmdrunner._MdbRead(tid,original,STR(M2CLASS.ISABSTRACT))
        pPackage = self.cmdrunner._MdbRead(tid,original,STR(M2CLASS.PACKAGE))
        createArgNames = [STR(M2CLASS.NAME), STR(M2CLASS.ISABSTRACT), STR(M2CLASS.PACKAGE)]
        createArgs = [pName, pIsabstract, pPackage]
        return (createArgNames,createArgs)

    def __GetCreateArgsM2Attribute(self, tid:int, original:Obj)->Tuple[List[STR],List[VAL]]:
        pName = self.cmdrunner._MdbRead(tid,original,STR(M2ATTRIBUTE.NAME))
        pClass = self.cmdrunner._MdbRead(tid,original,STR(M2ATTRIBUTE.CLASS))
        pPrimtype = self.cmdrunner._MdbRead(tid,original,STR(M2ATTRIBUTE.PRIMTYPE))
        pMul = self.cmdrunner._MdbRead(tid,original,STR(M2ATTRIBUTE.MUL))
        pUnit = self.cmdrunner._MdbRead(tid,original,STR(M2ATTRIBUTE.UNIT))
        pEnum = self.cmdrunner._MdbRead(tid,original,STR(M2ATTRIBUTE.ENUM))
        createArgNames = [STR(M2ATTRIBUTE.NAME), STR(M2ATTRIBUTE.CLASS), STR(M2ATTRIBUTE.PRIMTYPE), STR(M2ATTRIBUTE.MUL), STR(M2ATTRIBUTE.UNIT), STR(M2ATTRIBUTE.ENUM)]
        createArgs = [pName, pClass, pPrimtype, pMul, pUnit, pEnum]
        return (createArgNames,createArgs)

    def __GetCreateArgsM2Association(self, tid:int, original:Obj)->Tuple[List[STR],List[VAL]]:
        pSrcname = self.cmdrunner._MdbRead(tid,original,STR(M2ASSOCIATION.SRCNAME))
        pSrcclass = self.cmdrunner._MdbRead(tid,original,STR(M2ASSOCIATION.SRCCLASS))
        pSrcmul = self.cmdrunner._MdbRead(tid,original,STR(M2ASSOCIATION.SRCMUL))
        pDstname = self.cmdrunner._MdbRead(tid,original,STR(M2ASSOCIATION.DSTNAME))
        pDstclass = self.cmdrunner._MdbRead(tid,original,STR(M2ASSOCIATION.DSTCLASS))
        pDstmul = self.cmdrunner._MdbRead(tid,original,STR(M2ASSOCIATION.DSTMUL))
        pAnydst = self.cmdrunner._MdbRead(tid,original,STR(M2ASSOCIATION.ANYDST))
        createArgNames = [STR(M2ASSOCIATION.SRCNAME), STR(M2ASSOCIATION.SRCCLASS), STR(M2ASSOCIATION.SRCMUL), STR(M2ASSOCIATION.DSTNAME), STR(M2ASSOCIATION.DSTCLASS), STR(M2ASSOCIATION.DSTMUL), STR(M2ASSOCIATION.ANYDST)]
        createArgs = [pSrcname, pSrcclass, pSrcmul, pDstname, pDstclass, pDstmul, pAnydst]
        return (createArgNames,createArgs)

    def __GetCreateArgsM2Composition(self, tid:int, original:Obj)->Tuple[List[STR],List[VAL]]:
        pName = self.cmdrunner._MdbRead(tid,original,STR(M2COMPOSITION.NAME))
        pParentclass = self.cmdrunner._MdbRead(tid,original,STR(M2COMPOSITION.PARENTCLASS))
        pChildclass = self.cmdrunner._MdbRead(tid,original,STR(M2COMPOSITION.CHILDCLASS))
        pMulchild = self.cmdrunner._MdbRead(tid,original,STR(M2COMPOSITION.MULCHILD))
        pAnychild = self.cmdrunner._MdbRead(tid,original,STR(M2COMPOSITION.ANYCHILD))
        createArgNames = [STR(M2COMPOSITION.NAME), STR(M2COMPOSITION.PARENTCLASS), STR(M2COMPOSITION.CHILDCLASS), STR(M2COMPOSITION.MULCHILD), STR(M2COMPOSITION.ANYCHILD)]
        createArgs = [pName, pParentclass, pChildclass, pMulchild, pAnychild]
        return (createArgNames,createArgs)

    def __GetCreateArgsM2Inheritance(self, tid:int, original:Obj)->Tuple[List[STR],List[VAL]]:
        pSubclass = self.cmdrunner._MdbRead(tid,original,STR(M2INHERITANCE.SUBCLASS))
        pSuperclass = self.cmdrunner._MdbRead(tid,original,STR(M2INHERITANCE.SUPERCLASS))
        createArgNames = [STR(M2INHERITANCE.SUBCLASS), STR(M2INHERITANCE.SUPERCLASS)]
        createArgs = [pSubclass, pSuperclass]
        return (createArgNames,createArgs)

    def __GetCreateArgsM1Model(self, tid:int, original:Obj)->Tuple[List[STR],List[VAL]]:
        pM2Model = self.cmdrunner._MdbRead(tid,original,STR(M1MODEL.M2MODEL))
        pName = self.cmdrunner._MdbRead(tid,original,STR(M1MODEL.NAME))
        createArgNames = [STR(M1MODEL.M2MODEL), STR(M1MODEL.NAME)]
        createArgs = [pM2Model, pName]
        return (createArgNames,createArgs)

    def __GetCreateArgsM1Object(self, tid:int, original:Obj)->Tuple[List[STR],List[VAL]]:
        pM2Class = self.cmdrunner._MdbRead(tid,original,STR(M1OBJECT.M2CLASS))
        pModel = self.cmdrunner._MdbRead(tid,original,STR(M1OBJECT.MODEL))
        pName = self.cmdrunner._MdbRead(tid,original,STR(M1OBJECT.NAME))
        createArgNames = [STR(M1OBJECT.M2CLASS), STR(M1OBJECT.MODEL), STR(M1OBJECT.NAME)]
        createArgs = [pM2Class, pModel, pName]
        return (createArgNames,createArgs)

    def __GetCreateArgsM1Attribute(self, tid:int, original:Obj)->Tuple[List[STR],List[VAL]]:
        pM2Attribute = self.cmdrunner._MdbRead(tid,original,STR(M1ATTRIBUTE.M2ATTRIBUTE))
        pObject = self.cmdrunner._MdbRead(tid,original,STR(M1ATTRIBUTE.OBJECT))
        pValue = self.cmdrunner._MdbRead(tid,original,STR(M1ATTRIBUTE.VALUE))
        createArgNames = [STR(M1ATTRIBUTE.M2ATTRIBUTE), STR(M1ATTRIBUTE.OBJECT), STR(M1ATTRIBUTE.VALUE)]
        createArgs = [pM2Attribute, pObject, pValue]
        return (createArgNames,createArgs)

    def __GetCreateArgsM1Association(self, tid:int, original:Obj)->Tuple[List[STR],List[VAL]]:
        pM2Association = self.cmdrunner._MdbRead(tid,original,STR(M1ASSOCIATION.M2ASSOCIATION))
        pSrc = self.cmdrunner._MdbRead(tid,original,STR(M1ASSOCIATION.SRC))
        pDst = self.cmdrunner._MdbRead(tid,original,STR(M1ASSOCIATION.DST))
        createArgNames = [STR(M1ASSOCIATION.M2ASSOCIATION), STR(M1ASSOCIATION.SRC), STR(M1ASSOCIATION.DST)]
        createArgs = [pM2Association, pSrc, pDst]
        return (createArgNames,createArgs)

    def __GetCreateArgsM1Composition(self, tid:int, original:Obj)->Tuple[List[STR],List[VAL]]:
        pM2Composition = self.cmdrunner._MdbRead(tid,original,STR(M1COMPOSITION.M2COMPOSITION))
        pParent = self.cmdrunner._MdbRead(tid,original,STR(M1COMPOSITION.PARENT))
        pChild = self.cmdrunner._MdbRead(tid,original,STR(M1COMPOSITION.CHILD))
        createArgNames = [STR(M1COMPOSITION.M2COMPOSITION), STR(M1COMPOSITION.PARENT), STR(M1COMPOSITION.CHILD)]
        createArgs = [pM2Composition, pParent, pChild]
        return (createArgNames,createArgs)

    # ## COPY BASE PROPERTIES HANDELRS

    def __CopyBasePropertiesMxElement(self, tid:int, original:Obj, clone:Obj):
        documentation = self.cmdrunner._MdbRead(tid,original,STR(MXELEMENT.DOCUMENTATION))
        if(not documentation.IsNone()):
            self.__MdbUpdate(tid,clone,STR(MXELEMENT.DOCUMENTATION),documentation,I64(0))
        owner = self.cmdrunner._MdbRead(tid,original,STR(MXELEMENT.OWNER))
        if(not owner.IsNone()):
            self.__MdbUpdate(tid,clone,STR(MXELEMENT.OWNER),owner,I64(0))
        group = self.cmdrunner._MdbRead(tid,original,STR(MXELEMENT.GROUP))
        if(not group.IsNone()):
            self.__MdbUpdate(tid,clone,STR(MXELEMENT.GROUP),group,I64(0))
        permissions = self.cmdrunner._MdbRead(tid,original,STR(MXELEMENT.PERMISSIONS))
        for v in permissions:
            self.__MdbUpdate(tid,clone,STR(MXELEMENT.PERMISSIONS),permissions,I64(-1))

    def __CopyBasePropertiesMxConstraint(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesMxElement(tid,original,clone)

    def __CopyBasePropertiesM2Package(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesMxElement(tid,original,clone)

    def __CopyBasePropertiesM2Model(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesM2Package(tid,original,clone)

    def __CopyBasePropertiesM2Enum(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesMxElement(tid,original,clone)

    def __CopyBasePropertiesM2OptionOfEnum(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesMxElement(tid,original,clone)

    def __CopyBasePropertiesM2Class(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesMxElement(tid,original,clone)

    def __CopyBasePropertiesM2Attribute(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesMxElement(tid,original,clone)

    def __CopyBasePropertiesM2Association(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesMxElement(tid,original,clone)

    def __CopyBasePropertiesM2Composition(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesMxElement(tid,original,clone)

    def __CopyBasePropertiesM2Inheritance(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesMxElement(tid,original,clone)

    def __CopyBasePropertiesM1Model(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesMxElement(tid,original,clone)
        modelroot = self.cmdrunner._MdbRead(tid,original,STR(M1MODEL.MODELROOT))
        if(not modelroot.IsNone()):
            self.__MdbUpdate(tid,clone,STR(M1MODEL.MODELROOT),modelroot,I64(0))

    def __CopyBasePropertiesM1Object(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesMxElement(tid,original,clone)

    def __CopyBasePropertiesM1Feature(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesMxElement(tid,original,clone)

    def __CopyBasePropertiesM1Attribute(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesM1Feature(tid,original,clone)

    def __CopyBasePropertiesM1Association(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesM1Feature(tid,original,clone)

    def __CopyBasePropertiesM1Composition(self, tid:int, original:Obj, clone:Obj):
        self.__CopyBasePropertiesM1Feature(tid,original,clone)




'''Permissions are indicate the what can be done with a model element. 
A permission is composed of three byte:
1. byte: owner (O) rights
2. byte: group (G) rights
3. byte: anybody (A) rights
Each byte has four bit flags, i.e. the CRUD -> CDRU flags

1 bit C: Create  (only makes sense for classes)
2 bit R: Read 
3 bit U: Update
4 bit D: Delete

permissions are stored as big endian integers or hex strings

Example:

- A class that can only be modified by the owner, but instantiated by the group and can be read by anybody


      owner  |  group  | anybody
     C R U D | C R U D | C R U D     
Bit: 1 1 1 1 | 1 1 0 0 | 0 1 0 0

Bjoern Annighoefer 2022
'''

from ..error import EOQ_ERROR_INVALID_VALUE
#type checking
from typing import Tuple

### PERMISSION ###

class PERMISSION_FLAGS:
    #ANYBODY
    ANYBODY_CREATE    = (2**3)
    ANYBODY_READ      = (2**2)
    ANYBODY_UPDADE    = (2**1)
    ANYBODY_DELETE    = (2**0)
    #GROUP
    GROUP_CREATE      = (2**3)<<4
    GROUP_READ        = (2**2)<<4
    GROUP_UPDADE      = (2**1)<<4
    GROUP_DELETE      = (2**0)<<4
    #OWNER
    OWNER_CREATE      = (2**3)<<8
    OWNER_READ        = (2**2)<<8
    OWNER_UPDADE      = (2**1)<<8
    OWNER_DELETE      = (2**0)<<8

def PermissionToStr(permission:int)->str:
    return "%3x"%(permission)

def StrToPermission(permissionStr:str)->int:
    return int(permissionStr,base=16)
 
### FEATURE PERMISSSIONS ###   
    
PERMISSION_STR_SEPERATOR = '::'

WILDCARD_FEATURE_NAME = '*'
    
def GenerateFeaturePermissionStr(featureName:str, permission:int)->str:
    return "%3x%s%s"%(permission,PERMISSION_STR_SEPERATOR,featureName) 

def ParseFeaturePermissionStr(featurePermissionStr:str)->Tuple[str,int]:
    i = featurePermissionStr.find(PERMISSION_STR_SEPERATOR)
    if(i != 3):
        raise EOQ_ERROR_INVALID_VALUE("Invalid feature permission string: %s"%(featurePermissionStr))
    permission = StrToPermission(featurePermissionStr[0:3])
    featureName = featurePermissionStr[5:]
    return (featureName,permission)
    



'''
 2022 Bjoern Annighoefer
'''

from .serialdomain import SerialDomain

from ..config import Config
from ..command import Cmd, Evt
from ..frame import Frm, FRM_TYPE
from ..serializer import DesCmd, CreateSerializer
from ..error import EOQ_ERROR_RUNTIME

from threading import Lock, Semaphore
#type checking
from typing import Tuple, Callable, List, Any


class Request:
    '''A class storing requests as long as a command is running
    '''
    def __init__(self, frmId:int):
        self.frmId = frmId
        self.signal = Semaphore(0)
        self.resStr = None #this can be used to transfare back the result to the sync thread
        self.serType = None
        
    def Complete(self, resStr:str, serType:str):
        '''Complete the request
        '''
        self.resStr = resStr
        self.serType = serType
        self.signal.release() #inform the waiting thread

class RemoteDomainClient(SerialDomain):
    '''A domain that forwards commands as frames via some serial interface'''
    def __init__(self, config:Config, name:str="RemoteDomainClient"):
        super().__init__(config)
        self.name = name #used for debugging purposes
        self.frmId = 0
        self.frmIdLock = Lock()
        #request Queue
        self.requestQueue = {} #the list of open requests
        self.requestQueueLock = Lock()
        
    #@Override
    def RawDo(self, cmd:Cmd, sessionId:str=None, readOnly=False):
        cmdStr = self.remoteCmdSerializer.SerCmd(cmd)
        (resStr,resSerType) = self.SerRawDo(cmdStr, sessionId, self.remoteCmdSerializer.Name(),readOnly)
        res = DesCmd(resSerType,resStr)
        return res
    
    #@Override
    def SerRawDo(self, cmdStr:str, sessionId:str=None, serType:str=None, readOnly=False)->Tuple[str,str]:
        #build the frame
        frmId = self.__GetNextFrmId()
        frm = Frm(FRM_TYPE.CMD, frmId, serType, cmdStr, sessionId, readOnly)
        request = Request(frmId)
        #add a new request to the stack
        self.requestQueueLock.acquire()
        self.requestQueue[frmId] = request
        self.requestQueueLock.release()
        #send the request
        frmStr = self.remoteFrmSerializer.SerFrm(frm)
        self.SendSerFrm(frmStr)
        #wait for the result
        gotResult = request.signal.acquire(timeout=self.config.connectTimeout)
        #remove the result from the expected ones, because nobody listens anymore.
        self.requestQueueLock.acquire()
        del self.requestQueue[frmId]
        self.requestQueueLock.release()
        #check if we got a result in time
        if(not gotResult):
            raise EOQ_ERROR_RUNTIME("%s: Receive timeout: disconnected?"%(self.name))
        resStr = request.resStr
        resSerType = request.serType
        return (resStr,resSerType)
    
    #@Override
    def Observe(self, callback:Callable[[List[Evt],Any,object],None], context:Any=None, sessionId:str=None)->None: #by default register for all events
        '''Forward the observe wish to the domain
        '''
        super().Observe(callback, context, sessionId)
        frm = Frm(FRM_TYPE.OBS, 0, "NON", "", sessionId)
        frmStr = self.remoteFrmSerializer.SerFrm(frm)
        self.SendSerFrm(frmStr)
    
    #@Override    
    def Unobserve(self, callback:Callable[[List[Evt],Any,object],None], context:Any=None, sessionId:str=None)->None:
        frm = Frm(FRM_TYPE.UBS, 0, "NON", "", sessionId)
        frmStr = self.remoteFrmSerializer.SerFrm(frm)
        self.SendSerFrm(frmStr)
        super().Unobserve(callback, context, sessionId)
         
    #@Override
    def IsEventDesired(self, evt, sourceSessionId:str, sessionId:str)->bool:
        return (sourceSessionId==sessionId)
    
    def OnSerFrmReceived(self, frmStr:str)->None:
        '''Needs to be called if frames were received
        '''
        frm = self.remoteFrmSerializer.DesFrm(frmStr)
        if(FRM_TYPE.RES == frm.typ):
            try:
                request = self.requestQueue[frm.uid]
                request.Complete(frm.dat, frm.ser)
            except KeyError:
                self.config.logger.Warn("%s: Skipped unexpected response: %s"%(self.name,frmStr))
        elif(FRM_TYPE.EVT == frm.typ):
            sourceSessionId = frm.sid
            if(self.config.enableRawEvents):    
                self.NotifyObservers(frm.dat, sourceSessionId)
            else:
                try:
                    evtCmd = DesCmd(frm.ser,frm.dat) #universal deserialization
                    evts = evtCmd.a
                    self.NotifyObservers(evts,sourceSessionId)
                except Exception as e:
                    self.config.logger.Error('%s: Event deserialization failed: %s'%(self.name,str(e)))
        else:
            self.config.logger.Warn("%s: Skipped unexpected frame type: %s"%(self.name,frm.typ))
            
    ### INTERFACE ###
    
    def SendSerFrm(self,frmStr:str)->None:
        '''This needs to be overwritten by the implementing domain '''
        raise NotImplemented()
    
    ### PRIVATE ###
        
    def __GetNextFrmId(self):
        self.frmIdLock.acquire()
        frmId = self.frmId;
        if(self.frmId < 2^32): #limit frame IDs to 4byte
            self.frmId += 1
        else: #wrap around
            self.frmId = 0
        self.frmIdLock.release()
        return frmId
    



'''
The interface of the class the does the user verification and access control on element level. 
Different implementations are possible as long as they comply to this interface

2022 Bjoern Annighoefer

'''

from ..verifier import Verifier

class AccessController(Verifier):
    def __init__(self):
        super().__init__()
    
    ### INITIAL ACCESS ###
    
    def AuthenticateUser(self, user:str, password:str)->bool:
        '''Returns true if user and password are valid
        '''
        raise NotImplemented()
    
    ### OBSERVATION ACCESS GUARDS ###
    
    def IsAllowedToObserve(self, user:str, eventType:str)->bool:
        raise NotImplemented()
    
    def IsAllowedToSuperobserve(self, user:str, eventType:str)->bool:
        raise NotImplemented()
    
    ### All other methods are inherited from Verifier ###



'''
 Bjoern Annighoefer 2019
'''

from ..frame import Frm
from ..command import Cmd
from ..query import Qry, Seg
from ..error import EOQ_ERROR_DOES_NOT_EXIST
#type checking
from typing import Type, Dict


class Serializer:
    ''' Base class for any serializer
    
    '''
    def Name(self)->str:
        '''Returns a three letter identifier for this kind of serialization 
        '''
        raise NotImplemented()
    
    ### SERIALIZE INTERFACES
    def Ser(self, val)->str:
        ''' Auto-selecting based on the content
        '''
        if(isinstance(val, Frm)):
            return self.SerFrm(val)
        elif(isinstance(val, Cmd)):
            return self.SerCmd(val)
        elif(isinstance(val, Qry)):
            return self.SerQry(val)
        elif(isinstance(val, Seg)):
            return self.SerSeg(val)
        else:
            return self.SerVal(val)
    
    def SerFrm(self, frm:Frm)->str:
        pass
    
    def SerCmd(self, cmd:Cmd)->str:
        pass
    
    def SerQry(self, qry:Qry)->str:
        pass
    
    def SerSeg(self, seg:Seg)->str:
        pass
    
    def SerVal(self, val):
        pass
    
    
    ### DESERIALIZE INTERFACES
    def Des(self, data):
        ''' Auto-selecting based on the content
        '''
        #stupid version,  just try the different parsers starting with the frame
        try: return self.DesFrm(data)
        except: pass
        try: return self.DesCmd(data)
        except: pass
        try: return self.DesQry(data)
        except: pass
        try: return self.DesSeg(data)
        except: pass
        try: return self.DesVal(data)
        except: pass
    
    def DesFrm(self, data) -> Frm:
        pass
    
    def DesCmd(self, data) -> Cmd:
        pass
    
    def DesQry(self, data) -> Qry:
        pass
    
    def DesSeg(self, data) -> Seg:
        pass
    
    def DesVal(self, data):
        pass
    
    
    
### SERIALIZER REGISTRY AND FACTORY ###
SERIALIZER_REGISTRY:Dict[str,Type[Serializer]] = {} #a list of serializer classes that can be instantiated
DEFAULT_SERIALIZERS:Dict[str,Serializer] = {} #a list of serializer instances that can be used

def RegisterSerializer(serializerClass:Type[Serializer]):
    serializer = serializerClass()
    name = serializer.Name()
    SERIALIZER_REGISTRY[name] = serializerClass
    DEFAULT_SERIALIZERS[name] = serializer

def CreateSerializer(name:str):
    try:
        return SERIALIZER_REGISTRY[name]()
    except KeyError:
        raise EOQ_ERROR_DOES_NOT_EXIST('Unknown serializer: %s'%(name))
    
def GetDefaultSerializer(name:str):
    try:
        return DEFAULT_SERIALIZERS[name]
    except KeyError:
        raise EOQ_ERROR_DOES_NOT_EXIST('Unknown serializer: %s'%(name))
    

# GLOBAL DESERIALIZATION FUNCTIONS ###    

def DesCmd(name:str, data:str) -> Cmd:
    '''Deserializes a command based on the serializer name'''
    serializer = GetDefaultSerializer(name)
    return serializer.DesCmd(data)


def SerCmd(name:str, cmd:Cmd) -> str:
    '''Serializes a command based on the serializer name'''
    serializer = GetDefaultSerializer(name)
    return serializer.SerCmd(cmd)






    
    

    


'''
 2022 Bjoern Annighoefer
'''

from .serialdomain import SerialDomain

from ..config import Config, EOQ_DEFAULT_CONFIG
from ..domain import Domain
from ..command import Cmd, IsReadOnlyCmd, Evt
from ..serializer import DesCmd
from ..error import EOQ_ERROR_INVALID_TYPE


from threading import Thread, Semaphore, Lock, Event
import sys #used to exit threads
from collections import deque
#type checking
from typing import Tuple, List, Callable, Any


class ThreadController():
    ''' A class to store a thread '''
    
    def __init__(self, shallRun:bool):
        self.shallRun = shallRun
        
    def ShallRun(self)->bool:
        return self.shallRun
    
    def Stop(self):
        self.shallRun = False
        
def StopableAcquire(lock:Lock, threadController:ThreadController):
    '''This tries to acquire the lock, but looks from time to time if the thread was canceled.
    '''
    gotLock = False
    while(not gotLock):
        gotLock = lock.acquire(timeout=1.0)
        if(not threadController.ShallRun()):
            sys.exit()


class Promise():
    def __init__(self):
        self.signal = Semaphore(0)
        self.v = None
        
    def Get(self, threadController:ThreadController):
        self.signal.acquire()
        return self.v
    
    def Set(self,v):
        self.v = v
        self.signal.release()


class Job:
    def __init__(self, id:int, cmd:Cmd, sessionId:str, readOnly:bool, nDomains:int, cmdStr:str=None, serType:str=None, serialCmdOnly:bool=False):
        self.id = id
        self.cmd = cmd
        self.sessionId = sessionId
        self.readOnly = readOnly
        self.cmdStr = cmdStr
        self.serType = serType
        self.serialCmdOnly = serialCmdOnly
        self.multiDomainJobSync = Event()
        self.resPromise = Promise()
        self.processedBy = [False for i in range(nDomains)] #domains that started with the job
        self.finishedBy = [False for i in range(nDomains)] #domains that have finished this job
        self.lock = Lock() #a private lock for this job
        self.isFinished = False # is set to true if 
        
    def Result(self):
        return self.resPromise #Warning: this is only a promise. You must wait for it

class DomainPool(SerialDomain):
    def __init__(self,domains:List[Domain], shallForwardSerializedCmds:bool=False, config:Config=EOQ_DEFAULT_CONFIG):
        super().__init__(config)
        self.domains = domains
        allSerialDomains = all([isinstance(d,SerialDomain) for d in domains])
        if(shallForwardSerializedCmds and not allSerialDomains):
            raise EOQ_ERROR_INVALID_TYPE("If serial forwarding is enabled, all domains must be SerialDomain.")
        self.shallForwardSerializedCmds = shallForwardSerializedCmds
        self.config = config
        self.nDomains = len(domains)
        self.jobCount = 0
        self.jobs = deque()
        self.jobsSignal = Semaphore(0)
        self.jobsLock = Lock()
        self.allDomainsReadyEvent = Event()
        self.currentJobLock = Lock()
        self.threadController = ThreadController(True)
        self.workers = []
        workerMethod = self.__WorkerThreadSer if self.shallForwardSerializedCmds else self.__WorkerThread
        for i in range(self.nDomains):
            worker = Thread(target=workerMethod, args=(i,))
            self.workers.append(worker)
            worker.start()
        
    def Close(self)->None:
        self.threadController.Stop()
        for w in self.workers:
            w.join()
        for d in self.domains:
            d.Close()
    
    #@Override
    def RawDo(self, cmd:Cmd, sessionId:str=None, readOnly:bool=False)->Cmd:
        #override readOnly, because it is essential for the performance
        readOnly = IsReadOnlyCmd(cmd)
        cmdStr = None
        if(self.shallForwardSerializedCmds):
            cmdStr = self.remoteCmdSerializer.SerCmd(cmd)
        job = self.__AddNewJob(cmd, cmdStr, self.remoteCmdSerializer.Name(), sessionId, readOnly, False)
        res = job.Result().Get(self.threadController) #blocks until the result is ready
        return res
    
    #@Override
    def SerRawDo(self, cmdStr:str, sessionId:str=None, serType:str=None, readOnly:bool=False)->Tuple[str,str]:
        job = self.__AddNewJob(None, cmdStr, serType, sessionId, readOnly, True)
        res = job.Result().Get(self.threadController) #blocks until the result is ready
        return res
    
    def __AddNewJob(self, cmd:Cmd, cmdStr:str, serType:str, sessionId:str, readOnly=False, serialCmdOnly:bool=False)->Job:
        jobId = self.jobCount
        self.jobCount += 1 #job count is not thread-safe, but currently there is no demand for that.
        job = Job(jobId,cmd,sessionId,readOnly,self.nDomains,cmdStr,serType,serialCmdOnly)
        self.jobsLock.acquire()
        self.jobs.appendleft(job)
        self.jobsLock.release()
        self.jobsSignal.release() #indicate that a new job is waiting
        return job
        
    #@Override
    def IsEventDesired(self, evt, sourceSessionId:str, sessionId:str)->bool:
        return (sourceSessionId==sessionId)
    
    def Observe(self, callback:Callable[[List[Evt],Any,object],None], context:Any=None, sessionId:str=None)->None:
        #Observe only domain 0. This is the masterdomain
        self.domains[0].Observe(callback, context=context, sessionId=sessionId)
        
    def Unobserve(self, callback:Callable[[List[Evt],Any,object],None], context:Any=None, sessionId:str=None)->None:
        self.domains[0].Unobserve(callback, context=context, sessionId=sessionId)
    
    def __WorkerThread(self, n:int):
        domain = self.domains[n] #this is the domain of this thread
        while(self.threadController.ShallRun()):
            job = self.__GetNextJob(n)
            if(None!=job): #job can be None if stopped by shallRun during __GetNextJob
                res = domain.RawDo(job.cmd,job.sessionId,job.readOnly)
                job.lock.acquire()
                job.finishedBy[n] = True
                if(job.readOnly or all(job.finishedBy) ):
                    job.resPromise.Set(res) #notify about results
                    job.isFinished = True
                #in case of multi domain jobs the workers must be synchronized
                if(not job.readOnly):
                    if(job.isFinished): #I was the last worker on this multi domain job, so close it.
                        self.jobsLock.acquire()
                        self.jobs.pop() 
                        self.jobsLock.release()
                        job.lock.release()
                        job.multiDomainJobSync.set()
                    else: # I am not the last one I have to wait
                        job.lock.release()
                        #print("W%d blocked"%(n))
                        job.multiDomainJobSync.wait() 
                #print("W%d: finished job %d"%(n,job.id))
                
    def __WorkerThreadSer(self, n:int):
        domain = self.domains[n] #this is the domain of this thread
        while(self.threadController.ShallRun()):
            job = self.__GetNextJob(n)
            if(None!=job): #job can be None if stopped by shallRun during __GetNextJob
                (resStr,serType) = domain.SerRawDo(job.cmdStr,job.sessionId,job.serType,job.readOnly)
                StopableAcquire(job.lock, self.threadController)
                job.finishedBy[n] = True
                if(job.readOnly or all(job.finishedBy) ):
                    if(job.serialCmdOnly):
                        job.resPromise.Set((resStr,serType))
                    else:
                        res = DesCmd(serType,resStr)
                        job.resPromise.Set(res) #notify about results
                    job.isFinished = True
                #in case of multi domain jobs the workers must be synchronized
                if(not job.readOnly):
                    if(job.isFinished): #I was the last worker on this multi domain job, so close it.
                        StopableAcquire(self.jobsLock, self.threadController)
                        self.jobs.pop() 
                        self.jobsLock.release()
                        job.lock.release()
                        job.multiDomainJobSync.set()
                    else: # I am not the last one I have to wait
                        job.lock.release()
                        #print("W%d blocked"%(n))
                        job.multiDomainJobSync.wait() 
                #print("W%d: finished job %d"%(n,job.id))
            
    def __GetNextJob(self, n:int)->Job:
        job = None
        while(None == job and self.threadController.ShallRun()):
            isNewJob = self.jobsSignal.acquire(timeout=1.0)
            if(isNewJob):
                StopableAcquire(self.jobsLock, self.threadController)
                nextJob = self.jobs[-1]
                #print("W%d: got job %d"%(n,nextJob.id))
                if(nextJob.readOnly):
                    job = nextJob
                    self.jobs.pop()
                else: #this job is for only this  domain
                    job = nextJob
                    StopableAcquire(job.lock, self.threadController)
                    job.processedBy[n] = True #mark as processed by this domain
                    if(not all(job.processedBy)):
                        self.jobsSignal.release() #make it possible for the remaining threads to get this job
                    job.lock.release()
                self.jobsLock.release()
        return job
                               
        



'''
2019 Bjoern Annighoefer
'''

from .logger import Logger

class NoLogger(Logger):
    '''A logger which does nothing
    '''
    def __init__(self):
        super().__init__()
        
    #@Override
    def ShallLog(self):
        return False
    
    


'''
This is an verifier implementation that groups multiple verifiers and executes each of their verification and notification functions.

2022 Bjoern Annighoefer

'''


from .verifier import Verifier

from ..value import STR, I64, PRM, VAL, NON, LST
from ..query import Obj
from ..mdb import ReadOnlyMdb

#type checking
from typing import List, Union

class MultiVerifier(Verifier):
    def __init__(self):
        super().__init__()
        self.verifiers:List[Verifier] = []
        
    def AddVerifier(self, v:Verifier)->None:
        self.verifiers.append(v)
        v.Init(self.mdb)
        
    def RemoveVerifier(self, v:Verifier)->None:
        self.verifiers.remove(v)

### INIT AND CLOSE ###
    
    #@Override
    def Init(self, mdb:ReadOnlyMdb)->None:
        super().Init(mdb)
        for v in self.verifiers:
            v.Init(mdb)
    
    #@Override    
    def Deinit(self):
        for v in self.verifiers:
            v.Deinit()
    
    ### PRE VERIFICATION ###
    
    #@Override
    def CreatePreVerify(self, classId:Union[STR,Obj], createArgs:LST, target:Obj, recoveryArgs:LST, user:str=None)->None:
        for v in self.verifiers:
            v.CreatePreVerify(classId, createArgs, target, recoveryArgs, user)
    
    #@Override
    def ReadPreVerify(self, target:Obj, featureName:STR, context:Obj=NON(), user:str=None)->None:
        for v in self.verifiers:
            v.ReadPreVerify(target, featureName, context, user)
    
    #@Override
    def UpdatePreVerify(self, target:Obj, featureName:STR, value:PRM, position:I64=I64(0), user:str=None)->None:
        for v in self.verifiers:
            v.UpdatePreVerify(target, featureName, value, position, user)
    
    #@Override
    def DeletePreVerify(self, target:Obj, user:str=None)->None:
        for v in self.verifiers:
            v.DeletePreVerify(target, user)
    
    ### POST VERIFICATION ###
    
    #@Override
    def CreatePostVerify(self, classId:Union[STR,Obj], createArgs:LST, target:Obj, recoveryArgs:LST, user:str=None)->None:
        for v in self.verifiers:
            v.CreatePostVerify(classId, createArgs, target, recoveryArgs, user)
    
    #@Override
    def UpdatePostVerify(self, target:Obj, featureName:STR, value:PRM, position:I64=I64(0), oldValue:PRM=NON(), user:str=None)->None:
        for v in self.verifiers:
            v.UpdatePostVerify(target, featureName, value, position, oldValue, user)
    
    #@Override
    def DeletePostVerify(self, target:Obj, classId:STR, createArgs:LST, recoveryArgs:LST, user:str=None)->None:
        for v in self.verifiers:
            v.DeletePostVerify(target, classId, createArgs, recoveryArgs, user)
    
    
    ### NOTIFIERS ####
    
    #@Override
    def CreateNotify(self, classId:Union[STR,Obj], createArgs:LST, target:Obj, recoveryArgs:LST, user:str=None)->None:
        for v in self.verifiers:
            v.CreateNotify(classId, createArgs, target, recoveryArgs, user)
    
    #@Override
    def ReadNotify(self, target:Obj, featureName:STR, context:Obj, val:VAL, user:str=None)->None:
        for v in self.verifiers:
            v.ReadNotify(target, featureName, context, val, user)
    
    #@Override
    def UpdateNotify(self, target:Obj, featureName:STR, value:PRM, position:I64, oldValue:PRM, user:str=None)->None:
        for v in self.verifiers:
            v.UpdateNotify(target, featureName, value, position, oldValue, user)
    
    #@Override
    def DeleteNotify(self, target:Obj, classId:STR, createArgs:LST, recoveryArgs:LST, user:str=None)->None:
        for v in self.verifiers:
            v.DeleteNotify(target, classId, createArgs, recoveryArgs, user)